from opendp.mod import Measure
import opendp.prelude as dp
from opendp._lib import import_optional_dependency


def pairwise_predict(data, x_cuts):
    """
    The function matches data points together into pairs,
    fits a line that passes through each pair of points,
    and then computes the y values of each line at `x_cuts`.
    That is, it predicts the y values at `x_cuts`, pairwise.
    """
    np = import_optional_dependency("numpy")
    # Get an even number of rows.
    data = np.array(data, copy=True)[: len(data) // 2 * 2]

    # Shuffling the data improves the utility of results.
    np.random.shuffle(data)

    # Split data into pairs, where pair i is (p1[i], p2[i]).
    p1, p2 = np.array_split(data, 2)

    # Compute differences.
    dx, dy = (p2 - p1).T

    # Compute the midpoints of the pairs.
    x_bar, y_bar = (p1 + p2).T / 2

    # Compute y values on the pairwise slopes at x_cuts.
    points = dy / dx * (x_cuts[None].T - x_bar) + y_bar

    # Pairs where the x difference is zero are degenerate.
    return points.T[dx != 0]


def make_pairwise_predict(x_cuts, runs: int = 1, T=float):
    """
    The parameter `runs` controls how many times randomized pairwise predictions are computed.
    The default is 1. Increasing `runs` can improve the robustness and accuracy of the results;
    however, it can also increase computational cost and amount of noise needed later in the algorithm.
    """
    np = import_optional_dependency("numpy")
    return dp.t.make_user_transformation(
        # Outputs are Nx2 float numpy arrays.
        input_domain=dp.numpy.array2_domain(num_columns=2, T=T),
        # Neighboring input datasets differ by addition/removal of rows.
        input_metric=dp.symmetric_distance(),
        # Outputs are Nx2 float numpy arrays, but are half as long.
        output_domain=dp.numpy.array2_domain(num_columns=2, T=T),
        # Neighboring output datasets also differ by additions/removals.
        output_metric=dp.symmetric_distance(),
        # Apply the function `runs` times.
        function=lambda x: np.vstack(
            [pairwise_predict(x, x_cuts) for _ in range(runs)]
        ),
        # Each execution of pairwise predict contributes b_in records.
        stability_map=lambda b_in: b_in * runs,
    )


def make_select_column(j, T=float):
    return dp.t.make_user_transformation(
        input_domain=dp.numpy.array2_domain(num_columns=2, T=T),
        input_metric=dp.symmetric_distance(),
        output_domain=dp.vector_domain(dp.atom_domain(T=T)),
        output_metric=dp.symmetric_distance(),
        function=lambda x: x[:, j],
        stability_map=lambda b_in: b_in,
    )


def make_private_percentile_medians(output_measure, y_bounds, scale, candidates_count=100):
    np = import_optional_dependency("numpy")
    # this median mechanism favors candidates closest to the true median
    m_median = dp.m.then_private_quantile(
        output_measure=output_measure,
        # Evenly spaced points between y_bounds
        candidates=np.linspace(*y_bounds, candidates_count),
        alpha=0.5,
        scale=scale,
    )
    # apply the median mechanism to the 25th and 75th percentile columns
    return dp.c.make_composition(
        [
            make_select_column(0) >> dp.t.then_drop_null() >> m_median,
            make_select_column(1) >> dp.t.then_drop_null() >> m_median,
        ]
    )


def make_private_theil_sen(
    output_measure: Measure,
    x_bounds: tuple[float, float],
    y_bounds: tuple[float, float],
    scale: float,
    runs: int = 1,
    candidates_count: int = 100,
    fraction_bounds: tuple[float, float] = (0.25, 0.75),
) -> dp.Measurement:
    """
    Makes a measurement that takes a numpy array of (x, y) pairs,
    and returns a (slope, intercept) tuple.

    >>> import numpy as np
    >>> meas = make_private_theil_sen(dp.max_divergence(), (0, 100), (0, 100), scale=1.0)
    >>> slope, intercept = meas(np.array([[x, x] for x in range(100)]))
    """
    np = import_optional_dependency("numpy")
    # x_cuts are the fraction_bounds (for example, 25th and 75th percentiles) of x_bounds.
    # We'll predict y's at these x_cuts.
    x_cuts = x_bounds[0] + (x_bounds[1] - x_bounds[0]) * np.array(fraction_bounds)

    # we want coefficients, not y values!
    # Luckily y values are related to coefficients via a linear system
    P_inv = np.linalg.inv(np.vstack([x_cuts, np.ones_like(x_cuts)]).T)

    return (
        # pairwise_predict will return a 2xN array of y values at the x_cuts.
        make_pairwise_predict(x_cuts, runs)
        # privately select median y values for each x_cut
        >> make_private_percentile_medians(
            output_measure, y_bounds, scale, candidates_count=candidates_count
        )
        # transform median y values to coefficients (slope and intercept)
        >> (lambda ys: P_inv @ ys)
    )



# type: ignore
def make_randomized_response(categories: set[T], prob: f64):
    input_domain = AtomDomain(bool)
    input_metric = DiscreteMetric()
    output_measure = MaxDivergence()

    categories = list(categories)

    if len(categories) < 2:  # |\label{line:num_cats}|
        raise ValueError("expected at least two categories")
    
    num_categories = len(categories)

    if not (1 / num_categories <= prob <= 1):  # |\label{line:range}|
        raise ValueError("probability must be within [1/num_categories, 1]")
    
    # prepare constant: |\label{line:map}|
    if prob == 1.0:
        c = float("inf")
    else:
        c = p.inf_div((1).neg_inf_sub(prob)) \
            .inf_mul(num_categories.inf_sub(1)) \
            .inf_ln()
    
    def privacy_map(d_in: u32) -> QO:
        if d_in == 0:
            return 0
        else: 
            return c

    def function(truth: bool) -> bool:  # |\label{line:fn}|
        index = categories.index(truth)
        sample = usize.sample_uniform_int_below(
            len(num_categories) - (0 if index == -1 else 1))
        
        if index != -1 and sample >= index:
            sample += 1
        
        lie = categories[sample]

        be_honest = sample_bernoulli_float(prob, false)
        is_member = index != -1
        return truth if be_honest and is_member else lie
    
    return Measurement(input_domain, function, input_metric, output_measure, privacy_map)


# type: ignore
def match_group_by_truncation(
    plan: DslPlan, identifier: Expr
) -> Optional[Tuple[DslPlan, Truncation, Bound]]:
    if not isinstance(plan, DslPlan.GroupBy): # |\label{line:check_groupby}|
        return None

    input = plan.input
    keys = plan.keys
    aggs = plan.aggs
    apply = plan.apply
    options = plan.options

    if apply is not None or options != GroupbyOptions.default():
        return None # |\label{line:check_apply}|

    ids, by = partition(lambda expr: expr == identifier, keys) # |\label{line:partition}|

    if not ids: # |\label{line:check_ids}|
        return None

    return (
        input,
        Truncation.GroupBy(keys, aggs),
        Bound(by=by, per_group=1, num_groups=None), # |\label{line:bound}|
    )


# part of Rust's standard lib, included for readability, with hardcoded types
def partition(
    predicate, iterable: Iterable[Expr]
) -> Tuple[HashSet[Expr], HashSet[Expr]]:
    true_set = set()
    false_set = set()
    for item in iterable:
        if predicate(item):
            true_set.add(item)
        else:
            false_set.add(item)
    return true_set, false_set



# type: ignore
def make_geometric(
    input_domain: DI,
    input_metric: MI,
    scale: f64,
    bounds: Option[tuple[DI_Atom, DI_Atom]],
) -> Measurement[DI, DI_Carrier, MI, MO]:
    input_space = input_domain, input_metric
    if bounds is None:
        return DiscreteLaplace(scale, k=None).make_noise(input_space)
    else:
        return ConstantTimeGeometric(scale, bounds).make_noise(input_space)



# type: ignore
def make_int_to_bigint_threshold(
    input_space: tuple[
        MapDomain[AtomDomain[TK], AtomDomain[TV]], L0PInfDistance[P, AbsoluteDistance[QI]]
    ],
) -> Transformation[
    MapDomain[AtomDomain[TK], AtomDomain[TV]],
    MapDomain[AtomDomain[TK], AtomDomain[IBig]],
    L0PInfDistance[P, AbsoluteDistance[QI]],
    L0PInfDistance[P, AbsoluteDistance[RBig]],
]:
    input_domain, input_metric = input_space

    def stability_map(d_in):
        l0, lp, li = d_in
        lp = UBig.try_from(lp)
        li = UBig.try_from(li)
        return l0, lp, li

    return Transformation.new(
        input_domain,
        MapDomain( # |\label{line:output-domain}|
            key_domain=input_domain.key_domain,
            value_domain=AtomDomain.default(IBig),
        ),
        Function.new(lambda x: {k: IBig.from_(v) for k, v in x.items()}), # |\label{line:function}|
        input_metric,
        L0PI.default(),
        StabilityMap.new_fallible(stability_map),
    )



# type: ignore
class CompositionMeasure(ApproximateMaxDivergence):
    def composability(  # |\label{line:composability}|
        self, adaptivity: Adaptivity
    ) -> Composability:
        if matches(adaptivity, Adaptivity.FullyAdaptive):
            raise "fully-adaptive composition is not currently supported for max-divergence"
        return Composability.Concurrent

    def compose(self, d_mids: Vec[Self_Distance]) -> Self_Distance:
        eps_g, del_g = 0.0, 0.0
        for eps_i, del_i in d_mids:
            eps_g = eps_g.inf_add(eps_i)
            del_g = del_g.inf_add(del_i)
        return eps_g, del_g



from doctest import run_docstring_examples, OPTIONFLAGS_BY_NAME

import pytest

import opendp.prelude as dp


def test_number_of_tests_found(request):
    tests_found = len(request.session.items)
    assert tests_found >= 835


def test_TODOs():
    '''
    The tests here do the bare minimum so the generated API has 100% test coverage:
    At least we know the functions run.
    If new methods are added to the API, and we don't have better tests elsewhere,
    we should add a minimal test hxere as a reminder, although
    the real tests for the FFI may be on the Rust side in many cases.
    '''
    # TODO: Add meaningful tests of measures:
    assert str(dp.smoothed_max_divergence()) == 'SmoothedMaxDivergence'

    # TODO: Add meaningful tests of partial constructors:
    dp.m.then_private_expr(None, None) # type: ignore[arg-type]
    dp.m.then_private_lazyframe(None, None) # type: ignore[arg-type]
    dp.t.then_is_equal(None)
    dp.t.then_metric_bounded()
    dp.t.then_ordered_random()
    dp.t.then_quantile_score_candidates([], 0.5)
    dp.t.then_sum_of_squared_deviations()


def capture_doctest_result(doctest, capsys, options={}):
    optionflags = sum(OPTIONFLAGS_BY_NAME[option] for option in options)
    run_docstring_examples(doctest, {}, optionflags=optionflags)
    return capsys.readouterr().out

def assert_pass(doctest, capsys, options={}):
    result = capture_doctest_result(doctest, capsys, options)
    assert 'Failed example' not in result

# Use caps-lock just so it's easier to read the tests and see the differences.
def assert_FAIL(doctest, capsys, options={}):
    result = capture_doctest_result(doctest, capsys, options)
    assert 'Failed example' in result


def test_doctest_ignore(capsys):
    assert_FAIL('>>> 1/0', capsys)
    assert_pass('>>> 1/0', capsys, {'SKIP'})
    assert_FAIL('>>> 1/0', capsys, {'IGNORE'})

    assert_FAIL('>>> 2+2', capsys)
    assert_pass('>>> 2+2', capsys, {'SKIP'})
    assert_FAIL('>>> 2+2', capsys, {'IGNORE'})

    assert_pass('>>> 2+2\n4', capsys)
    assert_pass('>>> 2+2\n4', capsys, {'SKIP'})
    assert_pass('>>> 2+2\n4', capsys, {'IGNORE'})

    assert_FAIL('>>> 2+2\nextra!', capsys)
    assert_pass('>>> 2+2\nextra!', capsys, {'SKIP'})
    assert_pass('>>> 2+2\nextra!', capsys, {'IGNORE'})

    with pytest.raises(Exception, match='IGNORE can not be used with other flags'):
        assert_pass('>>> 2+2', capsys, {'IGNORE', 'NUMBER'})


# Include "polars" in name to filter out in smoke-test.yml.
def test_doctest_fuzzy_polars(capsys):
    doctest_in = '''
    >>> import polars as pl
    >>> pl.DataFrame({'col': ['value']})'''
    doctest_out = '''
    shape: (1, 1)
    ┌───────┐
    │ col   │
    │ ---   │
    │ str   │
    ╞═══════╡
    │ value │
    └───────┘
    '''
    doctest_out_rows = '''
    shape: (2, 1)
    ┌───────────┐
    │ col       │
    │ ---       │
    │ str       │
    ╞═══════════╡
    │ value     │
    │ surprise! │
    └───────────┘
    '''
    doctest_out_cols = '''
    shape: (1, 100)
    ┌───────┐
    │ col   │
    │ ---   │
    │ str   │
    ╞═══════╡
    │ value │
    └───────┘
    '''
    assert_FAIL(doctest_in, capsys)
    assert_pass(doctest_in, capsys, {'SKIP'})
    assert_FAIL(doctest_in, capsys, {'FUZZY_DF'})

    assert_pass(doctest_in + doctest_out, capsys)
    assert_pass(doctest_in + doctest_out, capsys, {'SKIP'})
    assert_pass(doctest_in + doctest_out, capsys, {'FUZZY_DF'})

    assert_FAIL(doctest_in + doctest_out_rows, capsys)
    assert_pass(doctest_in + doctest_out_rows, capsys, {'SKIP'})
    assert_pass(doctest_in + doctest_out_rows, capsys, {'FUZZY_DF'})

    assert_FAIL(doctest_in + doctest_out_cols, capsys)
    assert_pass(doctest_in + doctest_out_cols, capsys, {'SKIP'})
    assert_FAIL(doctest_in + doctest_out_cols, capsys, {'FUZZY_DF'})
    
    with pytest.raises(Exception, match='FUZZY_DF can not be used with other flags'):
        assert_pass(doctest_in, capsys, {'FUZZY_DF', 'NUMBER'})


# type: ignore
def match_per_group_predicate(
    enumeration: Expr,
    partition_by: Vec[Expr],
    identifier: Expr,
    threshold: u32,
) -> Optional[Bound]:
    # reorderings of an enumeration are still enumerations
    if isinstance(enumeration, Expr.Function) and isinstance( # |\label{line:reordering_start}|
        enumeration.options.collect_groups, ApplyOptions.GroupWise
    ):
        input = enumeration.input
        function = enumeration.function

        # FunctionExprs that may reorder data
        if function == FunctionExpr.Reverse:
            is_reorder = True
        elif isinstance(function, FunctionExpr.Random):
            method = str(function.method)
            is_reorder = method == "shuffle"
        else:
            is_reorder = False
        
        if is_reorder:
            enumeration = input[0]

    elif isinstance(enumeration, Expr.SortBy):
        for key in enumeration.by:
            check_infallible(key, Resize.Ban)
        enumeration = enumeration.expr# |\label{line:reordering_end}|

    # in Rust, the != results in a boolean comparison, not a "ne" expression
    if enumeration != int_range(lit(0), len(partition_by), 1, DataType.Int64): # |\label{line:check_enumeration}|
        return None

    # we now know this is a per group predicate,
    # and can raise more informative error messages

    # check if the function is limiting partition contributions
    ids, by = partition(lambda expr: expr == identifier, partition_by) # |\label{line:partition}|

    if not ids:
        raise "failed to find identifier column in per_group predicate condition"

    return Bound(by=by, per_group=threshold, num_groups=None) # |\label{line:bound}|



# type: ignore
def make_private_group_by(
    input_domain: DslPlanDomain,
    input_metric: FrameDistance[MI],
    output_measure: MO,
    plan: DslPlan,
    global_scale: Optional[f64],
    threshold: Optional[u32],
):
    input, group_by, aggs, key_sanitizer = match_group_by(plan) # |\label{line:match-group-by}|

    # 1: establish stability of `group_by` |\label{line:input_stability}|
    t_prior = input.make_stable(input_domain, input_metric)  # |\label{line:tprior}|
    middle_domain, middle_metric = t_prior.output_space()

    for expr in group_by: # |\label{line:group-by-stability}|
        # grouping keys must be stable
        t_group_by = expr.make_stable(
            WildExprDomain(
                columns=middle_domain.series_domains,
                context=ExprContext.RowByRow,
            ),
            L0PInfDistance(middle_metric[0]),
        )

        series_domain = t_group_by.output_domain.column
        try:
            domain = series_domain.element_domain(CategoricalDomain)
        except Exception:
            pass
            
        if domain is not None and domain.categories() is None:
            raise "Categories are data-dependent, which may reveal sensitive record ordering."

    group_by_id = HashSet.from_iter(group_by)  # |\label{line:groupcols}|
    margin = middle_domain.get_margin(group_by_id)  # |\label{line:margin}|

    # 2: prepare for release of `aggs` |\label{line:prep-release-aggs}|
    match key_sanitizer:
        case KeySanitizer.Join(keys):
            num_keys = LazyFrame.from_(keys).select([len()]).collect()
            margin.max_num_partitions = num_keys.column("len").u32().last()  # |\label{line:keys-len}|
            is_join = True
        case _:
            is_join = False

    m_expr_aggs = [
        make_private_expr(
            WildExprDomain(  # |\label{line:joint}|
                columns=middle_domain.series_domains,
                context=ExprContext.Aggregation(margin),
            ),
            PartitionDistance(middle_metric),
            output_measure,
            expr,
            global_scale,
        ) for expr in aggs
    ]
    m_aggs = make_composition(m_expr_aggs)

    f_comp = m_aggs.function
    f_privacy_map = m_aggs.privacy_map

    # 3: prepare for release of `keys` |\label{line:prep-release-keys}|
    dp_exprs, null_exprs = zip(*((plan.expr, plan.fill) for plan in m_aggs.invoke(input)))

    # 3.2: reconcile information about the threshold |\label{line:reconcile-threshold}|
    if margin.invariant is not None or is_join:  # |\label{line:no-thresholding}|
        threshold_info = None
    elif match_filter(key_sanitizer) is not None:  # |\label{line:sanitizer-threshold}|
        name, threshold_value = match_filter(key_sanitizer)
        noise = find_len_expr(dp_exprs, name)[1]
        threshold_info = name, noise, threshold_value, False
    elif threshold is not None:  # |\label{line:constructor-threshold}|
        name, noise = find_len_expr(dp_exprs, None)
        threshold_info = name, noise, threshold, True
    else:  # |\label{line:not-private}|
        raise f"The key set of {group_by_id} is private and cannot be released."

    # 3.3: update key sanitizer |\label{line:final-sanitizer}|
    if threshold_info is not None:  # |\label{line:incorporate-threshold}|
        name, _, threshold_value, is_present = threshold_info
        threshold_expr = col(name).gt(lit(threshold_value))
        if not is_present and predicate is not None:  # |\label{line:new-threshold-sanitizer}|
            key_sanitizer = KeySanitizer.Filter(threshold_expr.and_(predicate))
        else:
            key_sanitizer = KeySanitizer.Filter(threshold_expr)

    elif isinstance(key_sanitizer, KeySanitizer.Join):  # |\label{line:incorporate-join-fill}|
        key_sanitizer.fill_null = []
        for dp_expr, null_expr in zip(dp_exprs, null_exprs):
            name = dp_expr.meta().output_name()
            if null_expr is None:
                raise f"fill expression for {name} is unknown"

            key_sanitizer.fill_null.append(col(name).fill_null(null_expr))
    
    # 4: build final measurement |\label{line:build-meas}|
    def function(arg: DslPlan) -> DslPlan:  # |\label{line:function}|
        output = DslPlan.GroupBy(
            input=arg,
            keys=group_by,
            aggs=[p.expr for p in f_comp.eval(arg)],
            apply=None,
            maintain_order=False,
        )
        match key_sanitizer:
            case KeySanitizer.Filter(predicate):
                output = DslPlan.Filter(input=output, predicate=predicate)
            case KeySanitizer.Join(
                how,
                left_on,
                right_on,
                options,
                fill_null,
                keys=labels,
            ):
                match how:  # |\label{line:join-how}|
                    case JoinType.Left:
                        input_left, input_right = labels, output
                    case JoinType.Right:
                        input_left, input_right = output, labels
                    case _:
                        raise "unreachable"

                output = DslPlan.HStack(
                    input=DslPlan.Join(
                        input_left,
                        input_right,
                        left_on,
                        right_on,
                        options,
                        predicates=[],
                    ),
                    exprs=fill_null,
                    options=ProjectionOptions.default(),
                )
        return output

    def privacy_map(d_in: Bounds):  # |\label{line:privacy-map}|
        bound = d_in.get_bound(group_by_id)

        # incorporate all information into optional bounds
        l0 = option_min(bound.num_groups, margin.max_groups)
        li = option_min(bound.per_group, margin.max_length)
        l1 = d_in.get_bound(HashSet.new()).per_group  # |\label{line:trivial-l1}|

        # reduce optional bounds to concrete bounds
        if l0 is not None and l1 is not None and li is not None:
            pass
        elif l1 is not None:
            l0 = l0 or l1 # |\label{line:l0-from-l1}|
            li = li or l1 # |\label{line:li-from-l1}|
        elif l0 is not None and li is not None:
            l1 = l0.inf_mul(li) # |\label{line:l1-from-l0-li}|
        else: # |\label{line: helpful_error_message}|            
            raise f"num_groups ({l0}), total contributions ({l1}), and per_group ({li}) are not sufficiently well-defined."

        d_out = f_privacy_map.eval((l0, l1, li))

        if margin.invariant is not None or is_join:  # |\label{line:privacy-map-static-keys}|
            pass
        elif threshold_info is not None:  # |\label{line:privacy-map-threshold}|
            _, noise, threshold_value, _ = threshold_info
            if li >= threshold_value:
                raise f"Threshold must be greater than {li}."
            
            d_instability = threshold_value.neg_inf_sub(li)
            delta_single = integrate_discrete_noise_tail(
                noise.distribution, noise.scale, d_instability
            )
            delta_joint = (1).inf_sub(
                (1).neg_inf_sub(delta_single).neg_inf_powi(IBig.from_(l0))
            )
            d_out = MO.add_delta(d_out, delta_joint)
        else:
            raise "the key-set is sensitive"

        return d_out

    return t_prior >> Measurement.new(
        middle_domain,
        function,
        middle_metric,
        output_measure,
        privacy_map,
    )



import opendp.prelude as dp
import pytest


def test_amplification():
    input_space = (
        dp.vector_domain(dp.atom_domain(bounds=(0.0, 10.0)), size=10),
        dp.symmetric_distance(),
    )
    meas = input_space >> dp.t.then_mean() >> dp.m.then_laplace(scale=0.5)

    amplified = dp.c.make_population_amplification(meas, population_size=100)
    print("amplified base laplace:", amplified([1.0] * 10))
    assert meas.check(2, 2.0 + 1e-6)
    assert not meas.check(2, 2.0)
    assert amplified.check(2, 1.494)
    assert not amplified.check(2, 0.494)


def test_fix_delta():
    input_space = dp.atom_domain(T=float, nan=False), dp.absolute_distance(T=float)

    m_gauss = dp.c.make_zCDP_to_approxDP(dp.m.make_gaussian(*input_space, 10.0))
    assert m_gauss.map(1.0).epsilon(1e-6) == 0.42994146883694934

    m_gauss_fixed = dp.c.make_fix_delta(m_gauss, 1e-6)
    assert m_gauss_fixed.map(1.0) == (0.42994146883694934, 1e-06)


def test_make_composition():
    input_space = dp.vector_domain(dp.atom_domain(T=int)), dp.symmetric_distance()
    composed = dp.c.make_composition(
        [
            input_space
            >> dp.t.then_count()
            >> dp.c.make_composition(
                [
                    dp.space_of(int) >> dp.m.then_laplace(scale=2.0),
                    dp.space_of(int) >> dp.m.then_laplace(scale=200.0),
                ]
            ),
            input_space
            >> dp.t.then_cast_default(bool)
            >> dp.t.then_cast_default(int)
            >> dp.t.then_count()
            >> dp.m.then_laplace(scale=2.0),
            input_space
            >> dp.t.then_cast_default(float)
            >> dp.t.then_impute_constant(0.0)
            >> dp.t.then_clamp((0.0, 10.0))
            >> dp.t.then_sum()
            >> dp.m.then_laplace(scale=2.0),
            dp.c.make_composition(
                [
                    input_space >> dp.t.then_count() >> dp.m.then_laplace(scale=2.0),
                    input_space
                    >> dp.t.then_count(TO=float)
                    >> dp.m.then_laplace(scale=2.0),
                    (
                        input_space
                        >> dp.t.then_cast_default(str)
                        >> dp.t.then_count_by_categories(categories=["0", "12", "22"])
                        >> dp.m.then_laplace(scale=2.0)
                    ),
                ]
            ),
        ]
    )

    print("Check:", composed.check(1, 2.0))
    print("Forward map:", composed.map(3))
    print("Invocation:", composed.invoke([22, 12]))


def test_deprecated_composition():
    input_space = dp.atom_domain(T=int), dp.absolute_distance(T=int)
    m_lap = dp.m.make_laplace(*input_space, 1.0)
    
    with pytest.deprecated_call():
        dp.c.make_basic_composition([m_lap] * 2)

    with pytest.deprecated_call():
        input_space >> dp.c.then_sequential_composition(
            m_lap.output_measure,
            1,
            [1.0] * 2,
        )


@pytest.mark.skip(reason="long-running process to detect potential memory leaks")
def test_make_composition_leak():

    # choose a vector-valued mechanism that should run quickly for large inputs
    # we want to add as little noise as possible, so that execution time is small
    space = dp.vector_domain(dp.atom_domain(T=int)), dp.l1_distance(T=int)
    meas = space >> dp.m.then_laplace(0.0)

    # memory usage remains the same when this line is commented,
    # supporting that AnyObject's free recursively frees children
    meas = dp.c.make_composition([meas])

    # watch for leaked AnyObjects with 1 million i32 values
    # memory would jump by ~40mb every iteration
    for i in range(1000):
        print("iteration", i)
        meas([0] * 1_000_000)


def test_make_composition_approx():
    input_space = dp.atom_domain(T=float, nan=False), dp.absolute_distance(T=float)
    composed_fixed = dp.c.make_composition(
        [
            dp.c.make_fix_delta(
                dp.c.make_zCDP_to_approxDP(dp.m.make_gaussian(*input_space, 1.0)), 1e-7
            )
        ]
        * 2
    )
    assert composed_fixed.map(1.0) == (11.341717670937543, 2e-07)


def test_cast_zcdp_approxdp():
    input_space = dp.atom_domain(T=float, nan=False), dp.absolute_distance(T=float)

    base_gaussian = input_space >> dp.m.then_gaussian(10.0, MO=dp.ZeroConcentratedDivergence)
    assert base_gaussian.map(1.0) == 0.005

    approx_gaussian = dp.c.make_zCDP_to_approxDP(base_gaussian)
    assert approx_gaussian.map(1.0).epsilon(1e-6) == 0.42994146883694934
    
def test_cast_azcdp_approxdp():
    m_azcdp = dp.m.make_user_measurement(
        dp.atom_domain(T=bool),
        dp.absolute_distance(T=float),
        dp.approximate(dp.zero_concentrated_divergence()),
        lambda x: x,
        # rho, and catastrophic delta
        lambda d_in: (d_in * 0.05, d_in * 1e-7),
    )
    m_asdp = dp.c.make_zCDP_to_approxDP(m_azcdp)
    curve, delta = m_asdp.map(1.0)
    assert delta == 1e-7

    m_adp = dp.c.make_fix_delta(m_asdp, delta=1e-6)
    assert m_adp.map(1.0) == (curve.epsilon(1e-6 - 1e-7), 1e-6)


def test_renyidp():
    import gc

    m_rdp = dp.m.make_user_measurement(
        dp.atom_domain(T=bool),
        dp.absolute_distance(T=float),
        dp.renyi_divergence(),
        lambda x: x,
        lambda d_in: (lambda alpha: d_in * alpha / 2.0),
    )
    rdp_curve = m_rdp.map(1.0)
    # this used to cause a use-after free,
    # as the ε(α) curve created by the map would not be kept alive
    gc.collect()
    assert rdp_curve(4.0) == 2.0


def test_make_approximate():
    input_space = dp.atom_domain(T=float, nan=False), dp.absolute_distance(T=float)

    # spot check that mechanisms get delta terms
    m_gauss = dp.c.make_approximate(dp.m.make_gaussian(*input_space, 1.0))
    assert m_gauss.map(1.0) == (0.5, 0.0)

    m_lap = dp.c.make_approximate(dp.m.make_laplace(*input_space, 1.0))
    assert m_lap.map(1.0) == (1.0, 0.0)

    # composition works properly
    m_comp = dp.c.make_composition(
        [
            m_lap,
            dp.c.make_fix_delta(
                dp.c.make_zCDP_to_approxDP(dp.m.make_gaussian(*input_space, 1.0)),
                delta=1e-6,
            ),
        ]
    )
    assert isinstance(m_comp.map(1.0), tuple)


def test_make_pureDP_to_zCDP():
    input_space = dp.atom_domain(T=float, nan=False), dp.absolute_distance(T=float)
    meas = dp.c.make_composition(
        [
            dp.c.make_pureDP_to_zCDP(dp.m.make_laplace(*input_space, 10.0)),
            dp.m.make_gaussian(*input_space, 10.0),
        ]
    )

    # (1/10)^2 / 2 + 1 / 10^2 / 2
    assert meas.map(1.0) == 0.010000000000000002


def test_make_fixed_approxDP_to_approxDP():
    input_space = dp.atom_domain(T=float, nan=False), dp.absolute_distance(T=float)
    fadp_meas = dp.c.make_approximate(dp.m.make_laplace(*input_space, 10.0))

    adp_meas = dp.c.make_fixed_approxDP_to_approxDP(fadp_meas)

    assert adp_meas.map(1.0).epsilon(delta=1e-7) == 0.1



# type: ignore
def peel_permute_and_flip(x: list[RBig], scale: RBig, k: usize):
    natural_order = []
    sorted_order = BTreeSet.new()

    for _ in range(min(k, x.len())):
        index = permute_and_flip(x, scale)  # `\label{permute-and-flip}`
        x.remove(index)  # `\label{remove-score}`

        # map index on modified x back to original x (postprocessing)
        for del_ in sorted_order:  # `|\label{postprocess}`
            if del_ <= index:
                index += 1
            else:
                break

        sorted_order.insert(index)
        natural_order.push(index)

    return natural_order



from opendp.extras.polars import Margin, Bound
import pytest
import opendp.prelude as dp
import warnings
import re
import os
import io
from datetime import date, time, datetime


def test_polars_version():
    pl = pytest.importorskip("polars")
    from opendp.mod import _EXPECTED_POLARS_VERSION

    assert pl.__version__ == _EXPECTED_POLARS_VERSION


def seed(schema):
    pl = pytest.importorskip("polars")
    return pl.DataFrame(None, schema, orient="row").lazy()  # type: ignore[attr-defined]


def example_series():
    pl = pytest.importorskip("polars")
    return [
        dp.series_domain("A", dp.option_domain(dp.atom_domain(T=dp.f64))),
        dp.series_domain("B", dp.atom_domain(T=dp.i32)),
        dp.series_domain("C", dp.option_domain(dp.atom_domain(T=dp.String))),
        dp.series_domain("D", dp.atom_domain(T=dp.i32)),
    ], [
        pl.Series("A", [1.0] * 50, dtype=pl.Float64),
        pl.Series("B", [1, 2, 3, 4, 5] * 10, dtype=pl.Int32),
        pl.Series("C", ["1"] * 49 + [None], dtype=pl.String),
        pl.Series("D", [2] * 50, dtype=pl.Int32),
    ]


def test_infer():
    pl = pytest.importorskip("polars")
    series = pl.Series("A", [1] * 100)
    assert dp.RuntimeType.infer(series) == dp.Series
    assert dp.RuntimeType.infer(pl.DataFrame(series)) == dp.DataFrame
    assert dp.RuntimeType.infer(pl.LazyFrame(series)) == dp.LazyFrame
    assert dp.RuntimeType.infer(pl.col("A")) == dp.Expr


def example_lf(margin=None, **kwargs):
    pl = pytest.importorskip("polars")
    domains, series = example_series()
    lf_domain, lf = dp.lazyframe_domain(domains), pl.LazyFrame(series)
    if margin is not None:
        margin = [col if isinstance(col, pl.Expr) else pl.col(col) for col in margin]
        lf_domain = dp.with_margin(lf_domain, Margin(by=margin, **kwargs))
    return lf_domain, lf


def test_expr_domain():
    series_domains, _ = example_series()

    pl = pytest.importorskip("polars")
    by = [pl.col("A"), pl.col("B")]
    dp.wild_expr_domain(series_domains, dp.polars.Margin(by=by))
    dp.wild_expr_domain(series_domains, dp.polars.Margin(by=by, max_groups=10))
    dp.wild_expr_domain(series_domains, dp.polars.Margin(by=by, invariant="keys"))


def test_domains():
    example_lf(margin=["B"])
    example_lf(margin=["B"], invariant="keys", max_length=50)
    example_lf(
        margin=["B"],
        invariant="lengths",
        max_length=50,
        max_groups=3,
    )


# data loaders
@pytest.mark.parametrize("domain,series", zip(*example_series()))
def test_series_ffi(domain, series):
    """ensure that series can be passed to/from Rust"""
    pl_testing = pytest.importorskip("polars.testing")

    t_ident = (domain, dp.symmetric_distance()) >> dp.t.then_identity()
    pl_testing.assert_series_equal(t_ident(series), series)


def test_lazyframe_ffi():
    """ensure that lazyframes can be passed to/from Rust"""
    pl_testing = pytest.importorskip("polars.testing")
    lf_domain, lf = example_lf()
    t_ident = (lf_domain, dp.symmetric_distance()) >> dp.t.then_identity()

    pl_testing.assert_frame_equal(t_ident(lf).collect(), lf.collect())


@pytest.mark.parametrize(
    "measure", [dp.max_divergence(), dp.zero_concentrated_divergence()]
)
def test_private_lazyframe_explicit_sum(measure):
    pl = pytest.importorskip("polars")
    pl_testing = pytest.importorskip("polars.testing")

    lf_domain, lf = example_lf(
        margin=["B"],
        invariant="keys",
        max_length=50,
        max_groups=10,
    )

    expr = pl.col("A").fill_nan(0.0).fill_null(0.0).clip(0.0, 1.0).sum().dp.noise(0.0)
    plan = seed(lf.collect_schema()).group_by("B").agg(expr)
    m_lf = dp.m.make_private_lazyframe(
        lf_domain, dp.symmetric_distance(), measure, plan, 0.0
    )

    df_exp = pl.DataFrame(
        [
            pl.Series("B", list(range(1, 6)), dtype=pl.Int32),
            pl.Series("A", [10.0] * 5),
        ]
    )
    df_act = m_lf(lf).collect()
    pl_testing.assert_frame_equal(df_act.sort("B"), df_exp)


@pytest.mark.parametrize(
    "measure", [dp.max_divergence(), dp.zero_concentrated_divergence()]
)
def test_private_lazyframe_sum(measure):
    pl = pytest.importorskip("polars")
    pl_testing = pytest.importorskip("polars.testing")

    lf_domain, lf = example_lf(
        margin=["B"],
        invariant="keys",
        max_length=50,
        max_groups=10,
    )
    expr = pl.col("A").fill_nan(0.0).fill_null(0.0).dp.sum((1.0, 2.0), scale=0.0)
    plan = seed(lf.collect_schema()).group_by("B").agg(expr)
    m_lf = dp.m.make_private_lazyframe(
        lf_domain, dp.symmetric_distance(), measure, plan, 0.0
    )

    expect = pl.DataFrame(
        [
            pl.Series("B", [1, 2, 3, 4, 5], dtype=pl.Int32),
            pl.Series("A", [10.0] * 5, dtype=pl.Float64),
        ]
    )
    pl_testing.assert_frame_equal(m_lf(lf).collect().sort("B"), expect)


@pytest.mark.parametrize(
    "measure", [dp.max_divergence(), dp.zero_concentrated_divergence()]
)
def test_private_lazyframe_mean(measure):
    pl = pytest.importorskip("polars")
    pl_testing = pytest.importorskip("polars.testing")

    lf_domain, lf = example_lf(
        margin=["B"],
        invariant="lengths",
        max_length=50,
        max_groups=10,
    )

    expr = pl.col("A").fill_nan(0.0).fill_null(0.0).dp.mean((1.0, 2.0), scale=(0.0, 0.0))
    plan = seed(lf.collect_schema()).group_by("B").agg(expr)
    m_lf = dp.m.make_private_lazyframe(
        lf_domain, dp.symmetric_distance(), measure, plan, 1.0
    )

    expect = pl.DataFrame(
        [
            pl.Series("B", [1, 2, 3, 4, 5], dtype=pl.Int32),
            pl.Series("A", [1.0] * 5, dtype=pl.Float64),
        ]
    )
    pl_testing.assert_frame_equal(m_lf(lf).collect().sort("B"), expect)


def test_cast():
    pl = pytest.importorskip("polars")
    lf_domain, lf = example_lf()
    m_lf = dp.t.make_stable_lazyframe(
        lf_domain,
        dp.symmetric_distance(),
        lf.select(pl.col("A").cast(int)),
    )

    assert m_lf(lf).collect()["A"].dtype == pl.Int64


def test_stable_expr():
    domain = dp.wild_expr_domain([dp.series_domain("A", dp.atom_domain(T=float))])
    dp.t.make_stable_expr(domain, dp.frame_distance(dp.symmetric_distance()), "A")


def test_private_expr():
    pl = pytest.importorskip("polars")
    pl_testing = pytest.importorskip("polars.testing")
    m_len = dp.m.make_private_expr(
        dp.wild_expr_domain([], dp.polars.Margin(by=[])),
        dp.l01inf_distance(dp.symmetric_distance()),
        dp.max_divergence(),
        dp.len(scale=1.0),
    )

    e_plan = m_len(pl.LazyFrame(dict()))

    pl_testing.assert_frame_equal(e_plan.plan, pl.LazyFrame(dict()))

    assert re.match("len().*:noise_plugin()", str(e_plan.expr))
    assert re.match("0.*:noise_plugin()", str(e_plan.fill))


def test_private_lazyframe_median():
    pl = pytest.importorskip("polars")
    pl_testing = pytest.importorskip("polars.testing")

    lf_domain, lf = example_lf(
        margin=["A"], invariant="keys", max_length=50
    )
    candidates = list(range(1, 6))
    expr = pl.col("B").dp.median(candidates, 1.0)
    plan = seed(lf.collect_schema()).group_by("A").agg(expr)
    m_lf = dp.m.make_private_lazyframe(
        lf_domain,
        dp.symmetric_distance(),
        dp.max_divergence(),
        plan,
        0.0,
    )
    expect = pl.DataFrame(
        [pl.Series("A", [1.0], dtype=pl.Float64), pl.Series("B", [3], dtype=pl.Int64)]
    )

    pl_testing.assert_frame_equal(m_lf(lf).collect(), expect)


@pytest.mark.parametrize(
    "measure", [dp.max_divergence(), dp.zero_concentrated_divergence()]
)
def test_filter(measure):
    """ensure that expr domain's carrier type can be passed to/from Rust"""
    pl = pytest.importorskip("polars")
    pl_testing = pytest.importorskip("polars.testing")

    lf_domain, lf = example_lf(margin=[], invariant="keys", max_length=50)

    plan = lf.filter(pl.col("B") < 2).select(dp.len(scale=0.0))

    m_lf = dp.m.make_private_lazyframe(
        lf_domain, dp.symmetric_distance(), measure, plan
    )

    expect = pl.DataFrame([pl.Series("len", [10], dtype=pl.UInt32)])
    pl_testing.assert_frame_equal(m_lf(lf).collect(), expect)


def test_onceframe_multi_collect():
    lf_domain, lf = example_lf()
    plan = seed(lf.collect_schema()).select(dp.len(0.0))
    m_lf = dp.m.make_private_lazyframe(
        lf_domain, dp.symmetric_distance(), dp.max_divergence(), plan
    )

    of = m_lf(lf)
    of.collect()
    with pytest.raises(dp.OpenDPException):
        of.collect()


def test_onceframe_lazy():
    pl = pytest.importorskip("polars")

    lf_domain, lf = example_lf()
    plan = seed(lf.collect_schema()).select(dp.len(0.0))
    m_lf = dp.m.make_private_lazyframe(
        lf_domain, dp.symmetric_distance(), dp.max_divergence(), plan
    )

    of = m_lf(lf)
    assert isinstance(of.lazy(), pl.LazyFrame)


@pytest.mark.parametrize(
    "measure", [dp.max_divergence(), dp.zero_concentrated_divergence()]
)
def test_mechanisms(measure):
    pl_testing = pytest.importorskip("polars.testing")

    pl = pytest.importorskip("polars")

    lf_domain, lf = example_lf()

    if measure == dp.max_divergence():
        expr = pl.len().dp.laplace(0.0)
    else:
        expr = pl.len().dp.gaussian(0.0)

    plan = seed(lf.collect_schema()).select(expr)
    m_lf = dp.m.make_private_lazyframe(
        lf_domain, dp.symmetric_distance(), measure, plan, 0.0
    )

    expect = pl.DataFrame([pl.Series("len", [50], dtype=pl.UInt32)])
    pl_testing.assert_frame_equal(m_lf(lf).collect(), expect)


def test_wrong_mechanism():
    pl = pytest.importorskip("polars")

    lf_domain, lf = example_lf()

    plan = seed(lf.collect_schema()).select(pl.len().dp.gaussian(0.0))
    with pytest.raises(dp.OpenDPException) as err:
        dp.m.make_private_lazyframe(
            lf_domain,
            dp.symmetric_distance(),
            dp.max_divergence(),
            plan,
            0.0,
        )
    assert "expected Laplace distribution, found Gaussian" in (err.value.message or "")


def test_polars_context():
    pl = pytest.importorskip("polars")

    lf = pl.LazyFrame(
        {"A": [1, 2, 3, 4], "B": ["x", "x", "y", None]},
        schema={"A": pl.Int32, "B": pl.String},
    )

    context = dp.Context.compositor(
        data=lf,
        privacy_unit=dp.unit_of(contributions=1),
        privacy_loss=dp.loss_of(epsilon=1.0),
        split_evenly_over=2,
        margins=[
            dp.polars.Margin(max_length=5),
            dp.polars.Margin(by=["B"], invariant="keys"),
        ],
    )

    (
        context.query()
        .with_columns(pl.col("B").is_null().alias("B_nulls"))
        .filter(pl.col("B_nulls"))
        .select(pl.col("A").fill_null(2.0).dp.sum((0, 3)))
        .release()
        .collect()
    )

    (
        context.query()
        .group_by("B")
        .agg(dp.len(), pl.col("A").fill_null(2).dp.sum((0, 3)))
        .release()
        .collect()
    )


def test_polars_describe():
    pl = pytest.importorskip("polars")
    pl_testing = pytest.importorskip("polars.testing")

    lf = pl.LazyFrame(schema={"A": pl.Int32, "B": pl.String})

    context = dp.Context.compositor(
        data=lf,
        privacy_unit=dp.unit_of(contributions=1),
        privacy_loss=dp.loss_of(epsilon=1.0),
        split_evenly_over=2,
        margins=[
            dp.polars.Margin(by=["B"], invariant="keys", max_length=5),
        ],
    )

    expected = pl.DataFrame(
        {
            "column": ["len", "A", "B"],
            "aggregate": ["Frame Length", "Sum", "Sum"],
            "distribution": ["Integer Laplace", "Integer Laplace", "Integer Laplace"],
            # * sensitivity of the count is 1 (adding/removing one row changes the count by at most one),
            # * sensitivity of each sum is 3 (adding/removing one row with value as big as three...)
            # Therefore the noise scale of the sum query should be 3x greater
            # in order to consume the same amount of budget as the count.
            "scale": [6.0, 18.0, 18.0],
        }
    )

    summer = pl.col("A").fill_null(2).dp.sum((0, 3))

    query = context.query().group_by("B").agg(dp.len(), summer, summer.alias("B"))

    actual = query.summarize()
    pl_testing.assert_frame_equal(expected, actual)

    accuracy = [
        dp.discrete_laplacian_scale_to_accuracy(6.0, 0.05),
        dp.discrete_laplacian_scale_to_accuracy(18.0, 0.05),
        dp.discrete_laplacian_scale_to_accuracy(18.0, 0.05),
    ]
    expected = expected.hstack([pl.Series("accuracy", accuracy)])
    actual = query.summarize(alpha=0.05)
    pl_testing.assert_frame_equal(expected, actual)


def test_polars_accuracy_threshold():
    pl = pytest.importorskip("polars")
    pl_testing = pytest.importorskip("polars.testing")

    context = dp.Context.compositor(
        data=pl.LazyFrame(schema={"A": pl.Int32, "B": pl.String}),
        privacy_unit=dp.unit_of(contributions=1),
        privacy_loss=dp.loss_of(epsilon=1.0, delta=1e-7),
        split_evenly_over=2,
        margins=[
            dp.polars.Margin(by=["B"], max_length=5),
        ],
    )

    expected = pl.DataFrame(
        {
            "column": ["len", "A"],
            "aggregate": ["Frame Length", "Sum"],
            "distribution": ["Integer Laplace", "Integer Laplace"],
            "scale": [4.000000000000001, 12.000000000000004],
            "threshold": [65, None],
        },
        schema_overrides={"threshold": pl.UInt32},
    )

    query = (
        context.query()
        .group_by("B")
        .agg(dp.len(), pl.col("A").fill_null(2).dp.sum((0, 3)))
    )

    actual = query.summarize()
    pl_testing.assert_frame_equal(expected, actual)


def test_polars_non_wrapping():
    pl = pytest.importorskip("polars")

    context = dp.Context.compositor(
        data=pl.LazyFrame({"A": ["x", "x", "y", None]}),
        privacy_unit=dp.unit_of(contributions=1),
        privacy_loss=dp.loss_of(epsilon=1.0),
        split_evenly_over=1,
    )
    # only calls that return a LazyFrame or LazyGroupBy are wrapped
    assert context.query().explain() == 'DF ["A"]; PROJECT */1 COLUMNS'
    assert context.query().collect_schema() == {"A": pl.String}

    # for coverage: attribute access works properly
    context.query()._ldf

    # serialize/deserialize roundtrip
    query = context.query().select(dp.len())
    serde_plan = pl.LazyFrame.deserialize(io.BytesIO(query.serialize()))  # type: ignore
    assert query.serialize() == serde_plan.serialize()


def test_polars_collect_early():
    pl = pytest.importorskip("polars")

    context = dp.Context.compositor(
        data=pl.LazyFrame({"A": ["x", "x", "y", None]}),
        privacy_unit=dp.unit_of(contributions=1),
        privacy_loss=dp.loss_of(epsilon=1.0),
        split_evenly_over=1,
    )

    # catch the user trying to collect before release
    with pytest.raises(ValueError):
        context.query().collect()

    with pytest.raises(ValueError):
        context.query().describe()


def test_polars_threshold_epsilon():
    pl = pytest.importorskip("polars")
    pl_testing = pytest.importorskip("polars.testing")

    lf = pl.LazyFrame(
        {"A": [1] * 1000, "B": ["x"] * 500 + ["y"] * 500},
        schema={"A": pl.Int32, "B": pl.String},
    )

    context = dp.Context.compositor(
        data=lf,
        privacy_unit=dp.unit_of(contributions=1),
        privacy_loss=dp.loss_of(epsilon=1.0, delta=1e-7),
        split_evenly_over=2,
        margins=[
            dp.polars.Margin(by=["A"], invariant="keys"),
        ],
    )

    actual = context.query().group_by("A").agg(dp.len()).summarize()

    expected = pl.DataFrame(
        {
            "column": ["len"],
            "aggregate": ["Frame Length"],
            "distribution": ["Integer Laplace"],
            "scale": [2.0],
        }
    )

    # check that no threshold is set when keys are known
    pl_testing.assert_frame_equal(actual, expected)

    # check that query runs.
    print('output should be two columns ("A" and "len") with one row (1, ~1000)')
    print(context.query().group_by("A").agg(dp.len()).release().collect())

    actual = context.query().group_by("B").agg(dp.len()).summarize()

    expected = pl.DataFrame(
        {
            "column": ["len"],
            "aggregate": ["Frame Length"],
            "distribution": ["Integer Laplace"],
            "scale": [2.0],
            "threshold": [33],
        },
        schema_overrides={"threshold": pl.UInt32},
    )

    # threshold should work out to 33
    pl_testing.assert_frame_equal(actual, expected)

    # check that query runs.
    print('output should be two columns ("B" and "len") with two rows (1, ~500) each')
    print(context.query().group_by("B").agg(dp.len()).release().collect())


def test_polars_threshold_rho():
    pl = pytest.importorskip("polars")
    pl_testing = pytest.importorskip("polars.testing")

    lf = pl.LazyFrame(
        {"A": [1] * 1000, "B": ["x"] * 500 + ["y"] * 500},
        schema={"A": pl.Int32, "B": pl.String},
    )

    context = dp.Context.compositor(
        data=lf,
        privacy_unit=dp.unit_of(contributions=1),
        privacy_loss=dp.loss_of(rho=0.5, delta=1e-7),
        split_evenly_over=2,
    )

    query = context.query().group_by("B").agg(dp.len())

    actual = query.summarize()
    expected = pl.DataFrame(
        {
            "column": ["len"],
            "aggregate": ["Frame Length"],
            "distribution": ["Integer Gaussian"],
            # rho = .5, split over two queries, so rho_0 = 0.25.
            # gaussian formula is rho_0 = (d_in / scale)^2 / 2, now solve for scale:
            # scale = 1 / sqrt(2 * rho_0) = sqrt(2) ~= 1.414
            "scale": [1.4142135623730954],
            # probability of a partition with a unique individual being present in outcome must be at most 1e-7
            # since d_in = 1, only at most one partition may change
            # (when d_in is greater, then must consider union bound over up to L0 partitions)
            #
            # probability of returning an unstable partition
            # is the probability that noise added to the count in the unstable partition exceeds the threshold
            # therefore we must limit probability of sampling noise values greater than t to at most delta
            #
            # mass of discrete gaussian tail greater than t is bounded above by mass of continuous gaussian tail gte t
            # mass of continuous gaussian tail gte t is:
            # erfc(t / scale / sqrt(2)) / 2
            #
            # if you let t = 7, then mass is 3.7e-7 (too heavy by a factor of ~3.7)
            # if you let t = 8, then mass is 7.7e-9 (sufficiently unlikely enough)
            #
            # noise added to a true count d_in of 1 results in a threshold of 9
            "threshold": [9],
        },
        schema_overrides={"threshold": pl.UInt32},
    )

    # threshold should work out to 9
    pl_testing.assert_frame_equal(actual, expected)

    # check that query runs.
    print('output should be two columns ("B" and "len") with two rows (1, ~500) each')
    release = query.release().collect()
    assert release.columns == ["B", "len"]
    assert len(release) == 2


@pytest.mark.skipif(
    os.getenv("FORCE_TEST_REPLACE_BINARY_PATH") != "1",
    reason="setting OPENDP_POLARS_LIB_PATH interferes with the execution of other tests",
)
def test_replace_binary_path():
    import os

    pl = pytest.importorskip("polars")
    expr = dp.len(scale=1.0)

    # check that the library overwrites paths
    os.environ["OPENDP_POLARS_LIB_PATH"] = "testing!"

    m_expr = dp.m.make_private_expr(
        dp.wild_expr_domain(example_series()[0], dp.polars.Margin(by=[])),
        dp.l01inf_distance(dp.symmetric_distance()),
        dp.max_divergence(),
        expr,
    )
    assert "testing!" in str(m_expr(pl.LazyFrame(dict())).expr)

    # check that local paths in new expressions get overwritten
    os.environ["OPENDP_POLARS_LIB_PATH"] = __file__
    assert str(dp.len(scale=1.0)) == f"len().{__file__}:noise([null, dyn float: 1])"

    # cleanup
    del os.environ["OPENDP_POLARS_LIB_PATH"]


def test_pickle_bomb():
    pl = pytest.importorskip("polars")

    from polars._utils.parse import parse_into_list_of_expressions  # type: ignore[import-not-found]
    from polars._utils.wrap import wrap_expr  # type: ignore[import-not-found]
    from opendp._lib import lib_path
    import io
    import pickle

    # modified from https://intoli.com/blog/dangerous-pickles/
    poison_binary = b"c__builtin__\neval\n(V1 / 0\ntR."

    # poison_binary raises a ZeroDivisionError if it is ever unpickled
    with pytest.raises(ZeroDivisionError):
        pickle.loads(poison_binary)

    # craft an expression that contains a poisoned pickle binary

    py_exprs = parse_into_list_of_expressions(
        (pl.len(), pl.lit("Laplace"), pl.lit(0.7))
    )

    # Replicates parts of register_plugin_function from Polars,
    # to allow injection of the specially-crafted pickle binary
    bomb_expr = wrap_expr(
        pl.polars.register_plugin_function(
            plugin_path=str(lib_path),
            function_name="noise",
            args=py_exprs,
            kwargs=poison_binary,
            is_elementwise=True,
            input_wildcard_expansion=False,
            returns_scalar=False,
            cast_to_supertype=False,
            pass_name_to_apply=False,
            changes_length=False,
        )
    )

    # craft a lazyframe that contains a poisoned pickle binary
    lf_domain, lf = example_lf()
    bomb_lf = lf.select(bomb_expr)

    # ensure that ser/de round-trip of expression does not trigger pickle
    ser_expr = bomb_expr.meta.serialize()
    pl.Expr.deserialize(io.BytesIO(ser_expr))

    # ensure that ser/de round-trip of lazyframe does not trigger pickle
    ser_lf = bomb_lf.serialize()
    pl.LazyFrame.deserialize(io.BytesIO(ser_lf))

    # OpenDP explicitly rejects any pickled data it finds
    err_msg_re = "OpenDP does not allow pickled keyword arguments as they may enable remote code execution."
    with pytest.raises(dp.OpenDPException, match=err_msg_re):
        dp.m.make_private_expr(
            dp.wild_expr_domain(example_series()[0], dp.polars.Margin(by=[])),
            dp.l01inf_distance(dp.symmetric_distance()),
            dp.max_divergence(),
            bomb_expr,
        )

    with pytest.raises(dp.OpenDPException, match=err_msg_re):
        dp.m.make_private_lazyframe(
            lf_domain,
            dp.symmetric_distance(),
            dp.max_divergence(),
            bomb_lf,
        )


def test_cut():
    pl = pytest.importorskip("polars")
    pl_testing = pytest.importorskip("polars.testing")

    data = pl.LazyFrame({"x": [0.4, 1.0, 1.5, 2.0, 2.5, 3.0, 3.5]})
    by = [pl.col("x").cut([1.0, 2.0, 3.0]).to_physical()]
    with warnings.catch_warnings():
        context = dp.Context.compositor(
            data=data,
            privacy_unit=dp.unit_of(contributions=1),
            privacy_loss=dp.loss_of(epsilon=10000.0),
            split_evenly_over=1,
            margins=[dp.polars.Margin(by=by, invariant="keys")],
        )
    actual = context.query().group_by(*by).agg(dp.len()).release().collect().sort("x")
    expected = pl.DataFrame(
        {"x": [0, 1, 2, 3], "len": [2, 2, 2, 1]},
        schema={"x": pl.UInt32, "len": pl.UInt32},
    )

    pl_testing.assert_frame_equal(actual, expected)


@pytest.mark.xfail(raises=AssertionError)
def test_csv_bad_encoding_loading():
    # See https://github.com/opendp/opendp/issues/1976
    # If a CSV is misencoded, Polars loads an empty dataframe.
    # Since we tell users not to look at their data,
    # we may want to try harder to load the csv,
    # or give more information on failure.
    pl = pytest.importorskip("polars")
    pl_testing = pytest.importorskip("polars.testing")
    import tempfile

    name = "André"
    name_b = name.encode("iso-8859-1")  # Polars only handles 'utf-8'.

    with tempfile.NamedTemporaryFile(delete=False) as fp:
        # By default, would delete file on "close()";
        # With "delete=False", clean up when exiting "with" instead.
        fp.write(b"name\n" + name_b)
        fp.close()
        observed = pl.scan_csv(fp.name, encoding="utf8-lossy")
        expected = pl.LazyFrame(
            {"name": [name]},
            schema={"name": pl.String},
        )
        pl_testing.assert_frame_equal(observed, expected)


def test_categorical_domain_no_mapping():
    pl = pytest.importorskip("polars")

    lf = pl.LazyFrame(
        [pl.Series("A", ["Texas", "New York", None], dtype=pl.Categorical)]
    )

    # query should be rejected even when data is known to be non-null,
    # because Polars will still raise a warning if the fill value is not in the encoding
    for element_domain in [
        dp.option_domain(dp.categorical_domain()),
        dp.categorical_domain(),
    ]:
        lf_domain = dp.lazyframe_domain([dp.series_domain("A", element_domain)])
        assert str(lf_domain) == "FrameDomain(A: cat; margins=[])"
        match_msg = re.escape("fill_null cannot be applied to categorical data")
        with pytest.raises(dp.OpenDPException, match=match_msg):
            dp.t.make_stable_lazyframe(
                lf_domain,
                dp.symmetric_distance(),
                lf.with_columns(pl.col("A").fill_null("A")),
            )


@pytest.mark.parametrize("wrap_with_option", [False, True])
def test_categorical_domain_with_mapping(wrap_with_option):
    pl = pytest.importorskip("polars")

    categories = ["Texas", "New York"]
    keys = pl.LazyFrame([pl.Series("A", categories, dtype=pl.Categorical)])
    lf = pl.LazyFrame([pl.Series("A", categories + [None], dtype=pl.Categorical)])

    element_domain = dp.categorical_domain(categories)
    if wrap_with_option:
        element_domain = dp.option_domain(element_domain)

    lf_domain = dp.lazyframe_domain([dp.series_domain("A", element_domain)])
    assert str(lf_domain) == "FrameDomain(A: cat; margins=[])"
    # checks that categorical grouping keys can be released if encoding is public
    dp.m.make_private_lazyframe(
        lf_domain,
        dp.symmetric_distance(),
        dp.max_divergence(),
        lf.group_by("A").agg(dp.len()).join(keys, how="right", on=["A"]),
        global_scale=1.0,
    )


def test_categorical_context():
    pl = pytest.importorskip("polars")

    lf = pl.LazyFrame(
        {"A": [1] * 1000, "B": ["x"] * 500 + ["y"] * 500},
        schema_overrides={"A": pl.Int32, "B": pl.Categorical},
    )

    context = dp.Context.compositor(
        data=lf,
        privacy_unit=dp.unit_of(contributions=1),
        privacy_loss=dp.loss_of(epsilon=1.0, delta=1e-6),
        split_evenly_over=1,
    )

    with pytest.raises(dp.OpenDPException, match=r"Categories are data-dependent"):
        context.query().group_by("B").agg(dp.len()).release()

    # check that query runs.
    print('output should be two columns ("B" and "len") with two rows (1, ~500)')
    release = (
        context.query()
        .select(pl.col.B.cast(str))
        .group_by("B")
        .agg(dp.len())
        .release()
        .collect()
    )
    assert release.shape == (2, 2)


def test_to_physical_unordered():
    pl = pytest.importorskip("polars")
    lf_domain = dp.lazyframe_domain([dp.series_domain("A", dp.categorical_domain())])
    lf = pl.LazyFrame([pl.Series("A", ["Texas", "New York"], dtype=pl.Categorical)])

    # check that row ordering is protected
    with pytest.raises(dp.OpenDPException, match=re.escape("to_physical: to prevent")):
        dp.t.make_stable_lazyframe(
            lf_domain,
            dp.symmetric_distance(),
            lf.with_columns(pl.col("A").to_physical()),
        )


def test_float_sum_with_unlimited_reorderable_partitions():
    pl = pytest.importorskip("polars")
    lf_domain = dp.lazyframe_domain(
        [
            dp.series_domain("region", dp.atom_domain(T=dp.i64)),
            dp.series_domain("income", dp.atom_domain(T=dp.f64)),
        ]
    )
    lf_domain = dp.with_margin(
        lf_domain, dp.polars.Margin(by=[pl.col("region")], invariant="lengths", max_length=6)
    )

    from opendp.domains import _lazyframe_from_domain

    lf = _lazyframe_from_domain(lf_domain)

    # sum of income per region, add noise with scale of 1.0
    pl = pytest.importorskip("polars")
    plan = lf.group_by("region").agg(
        [pl.col("income").fill_nan(0.0).dp.sum(bounds=(1_000, 100_000), scale=1.0)]
    )

    # since there are an unknown number of partitions, and each partition has non-zero sensitivity, sensitivity is undefined
    with pytest.raises(
        dp.OpenDPException,
        match="max_groups must be known when the metric is not sensitive to ordering",
    ):
        dp.m.make_private_lazyframe(
            lf_domain, dp.symmetric_distance(), dp.max_divergence(), plan
        )


def test_sort_usability():
    pl = pytest.importorskip("polars")

    context = dp.Context.compositor(
        data=pl.LazyFrame(schema={"A": pl.Int32}),
        privacy_unit=dp.unit_of(contributions=1),
        privacy_loss=dp.loss_of(epsilon=1.0),
        split_evenly_over=1,
    )
    with pytest.raises(dp.OpenDPException, match="Found sort in query plan."):
        context.query().select(pl.len().dp.noise()).sort(by="A").release()


def test_count_queries():
    pl = pytest.importorskip("polars")
    pl_testing = pytest.importorskip("polars.testing")

    lf_domain = dp.lazyframe_domain(
        [dp.series_domain("data", dp.atom_domain(T=dp.i64))]
    )

    from opendp.domains import _lazyframe_from_domain

    lf = _lazyframe_from_domain(lf_domain)

    # sum of income per region, add noise with scale of 1.0
    plan = lf.select(
        [
            pl.col("data").dp.len(scale=0).alias("len"),
            pl.col("data").dp.count(scale=0).alias("count"),
            pl.col("data").dp.n_unique(scale=0).alias("n_unique"),
            pl.col("data").dp.null_count(scale=0).alias("null_count"),
        ]
    )

    m_counts = dp.m.make_private_lazyframe(
        lf_domain, dp.symmetric_distance(), dp.max_divergence(), plan
    )

    release = m_counts(pl.LazyFrame({"data": [1, 1, 1, None]})).collect()
    expected = pl.DataFrame(
        {
            "len": [4],
            "count": [3],
            "n_unique": [2],
            "null_count": [1],
        }
    ).cast({pl.Int64: pl.UInt32})
    pl_testing.assert_frame_equal(release, expected)


def test_explicit_grouping_keys():
    pl = pytest.importorskip("polars")
    pl_testing = pytest.importorskip("polars.testing")

    lf_domain, lf = example_lf(margin=["B"], max_length=100)

    plan_right = (
        seed(lf.collect_schema()).group_by("B").agg(pl.col("D").dp.sum((0, 10)))
    )
    labels = pl.LazyFrame(pl.Series("B", [2, 3, 4, 5, 6], dtype=pl.Int32))
    plan = labels.join(plan_right, on="B", how="left")
    m_lf = dp.m.make_private_lazyframe(
        lf_domain, dp.symmetric_distance(), dp.max_divergence(), plan, 0.0
    )

    df_act = m_lf(lf).collect()

    df_exp = pl.DataFrame(
        [
            pl.Series("B", list(range(2, 7)), dtype=pl.Int32),
            pl.Series("D", [20] * 4 + [0], dtype=pl.Int32),
        ]
    )
    pl_testing.assert_frame_equal(df_act.sort("B"), df_exp)


def test_explicit_grouping_keys_context():
    pl = pytest.importorskip("polars")
    pl_testing = pytest.importorskip("polars.testing")

    lf_domain, lf = example_lf(margin=["B"], max_length=100)

    context = dp.Context.compositor(
        data=lf,
        privacy_unit=dp.unit_of(contributions=1),
        privacy_loss=dp.loss_of(epsilon=1.0),
        split_evenly_over=1,
        domain=lf_domain,
    )

    keys = pl.DataFrame(pl.Series("B", [2, 3, 4, 5, 6], dtype=pl.Int32))
    query = (
        context.query().group_by("B").agg(pl.col("D").dp.sum((0, 10))).with_keys(keys)
    )
    observed = query.release().collect().sort("B")

    expected = pl.DataFrame(
        [
            pl.Series("B", list(range(2, 7)), dtype=pl.Int32),
            pl.Series("D", [20] * 4 + [0], dtype=pl.Int32),
        ]
    )
    assert observed.collect_schema() == expected.collect_schema()
    observed = observed.with_columns(D=expected["D"])
    pl_testing.assert_frame_equal(observed, expected)


@pytest.mark.parametrize("dtype", ["Time", "Datetime", "Date"])
def test_datetime(dtype):
    pl = pytest.importorskip("polars")
    dtype = getattr(pl, dtype)

    context = dp.Context.compositor(
        data=pl.LazyFrame([pl.Series("x", ["2000-01-02T03:04:05"])]),
        privacy_unit=dp.unit_of(contributions=1),
        privacy_loss=dp.loss_of(epsilon=1.0, delta=1e-7),
        split_evenly_over=1,
    )
    query = (
        context.query()
        .group_by(pl.col.x.str.strptime(format=r"%Y-%m-%dT%H:%M:%S", dtype=dtype))
        .agg(dp.len())
    )
    observed = query.release().collect()
    assert observed["x"].dtype == dtype


def test_temporal_domain():
    pl = pytest.importorskip("polars")

    # this triggers construction of a lazyframe domain from the schema
    context = dp.Context.compositor(
        data=pl.LazyFrame(
            [
                pl.Series("date", [date(2000, 1, 1)]),
                pl.Series("time", [time(7, 8, 9)]),
                pl.Series("datetime", [datetime(2000, 1, 1, 7, 8, 9)]),
            ]
        ),
        privacy_unit=dp.unit_of(contributions=1),
        privacy_loss=dp.loss_of(epsilon=1.0, delta=1e-7),
        split_evenly_over=1,
    )
    observed = context.accountant.input_domain
    expected = dp.lazyframe_domain(
        [
            dp.series_domain("date", dp.option_domain(dp.atom_domain(T="NaiveDate"))),
            dp.series_domain("time", dp.option_domain(dp.atom_domain(T="NaiveTime"))),
            dp.series_domain(
                "datetime", dp.option_domain(dp.datetime_domain(time_unit="us"))
            ),
        ]
    )

    # check that domain is as expected
    assert observed == expected


def test_replace():
    pl = pytest.importorskip("polars")
    pl_testing = pytest.importorskip("polars.testing")

    # this triggers construction of a lazyframe domain from the schema
    context = dp.Context.compositor(
        data=pl.LazyFrame(pl.Series("alpha", ["A", "B", "C"] * 100)),
        privacy_unit=dp.unit_of(contributions=1),
        privacy_loss=dp.loss_of(epsilon=1.0, delta=1e-7),
        split_evenly_over=4,
        margins=[dp.polars.Margin(by=(), max_length=300)],
    )

    # replace multiple input values with one output value
    pl_testing.assert_series_equal(
        context.query()
        .group_by(pl.col.alpha.replace(["A", "B"], "D"))
        .agg(dp.len())
        .release()
        .collect()["alpha"]
        .sort(),
        pl.Series("alpha", ["C", "D"]),
    )

    # replace multiple input values with respective output values (lists)
    pl_testing.assert_series_equal(
        context.query()
        .group_by(pl.col.alpha.replace(["A", "B"], ["D", "E"]))
        .agg(dp.len())
        .release()
        .collect()["alpha"]
        .sort(),
        pl.Series("alpha", ["C", "D", "E"]),
    )

    # replace multiple input values with respective output values (dict)
    pl_testing.assert_series_equal(
        context.query()
        .group_by(pl.col.alpha.replace({"A": "D", "B": "E"}))
        .agg(dp.len())
        .release()
        .collect()["alpha"]
        .sort(),
        pl.Series("alpha", ["C", "D", "E"]),
    )


def test_replace_strict():
    pl = pytest.importorskip("polars")

    # this triggers construction of a lazyframe domain from the schema
    context = dp.Context.compositor(
        data=pl.LazyFrame(pl.Series("alpha", ["A", "B", "C"] * 100)),
        privacy_unit=dp.unit_of(contributions=1),
        privacy_loss=dp.loss_of(epsilon=1.0, delta=1e-7),
        split_evenly_over=1,
        margins=[dp.polars.Margin(by=(), max_length=300)],
    )

    # replace multiple input values with one output value
    assert isinstance(
        context.query()
        .select(
            pl.col("alpha")
            .replace_strict(["A", "B"], [1, 2], default=0, return_dtype=pl.Int64)
            .dp.sum((0, 2))
        )
        .release()
        .collect()["alpha"][0],
        int,
    )


def test_cast_enum():
    pl = pytest.importorskip("polars")
    pl_testing = pytest.importorskip("polars.testing")
    # this triggers construction of a lazyframe domain from the schema
    context = dp.Context.compositor(
        data=pl.LazyFrame(pl.Series("alpha", ["A", "B", "C"] * 100)),
        privacy_unit=dp.unit_of(contributions=1),
        privacy_loss=dp.loss_of(epsilon=1.0, delta=1e-7),
        split_evenly_over=1,
        margins=[dp.polars.Margin(by=(), max_length=300)],
    )

    enum_dtype = pl.Enum(["A", "B", "C"])

    pl_testing.assert_series_equal(
        context.query()
        .with_columns(pl.col.alpha.cast(enum_dtype))
        .group_by(pl.col.alpha)
        .agg(dp.len())
        .release()
        .collect()["alpha"]
        .sort()
        .cast(enum_dtype),
        pl.Series("alpha", ["A", "B", "C"], enum_dtype),
    )


def test_enum_domain():
    pl = pytest.importorskip("polars")

    enum_dtype = pl.Enum(["A", "B", "C"])
    # this triggers construction of a lazyframe domain from the schema
    context = dp.Context.compositor(
        data=pl.LazyFrame(pl.Series("alpha", ["A", "B", "C"] * 100, dtype=enum_dtype)),
        privacy_unit=dp.unit_of(contributions=1),
        privacy_loss=dp.loss_of(epsilon=1.0, delta=1e-7),
        split_evenly_over=1,
    )
    observed = context.accountant.input_domain
    expected = dp.lazyframe_domain(
        [
            dp.series_domain(
                "alpha", dp.option_domain(dp.enum_domain(enum_dtype.categories))
            ),
        ]
    )

    # check that domain is as expected
    assert observed == expected


@pytest.mark.xfail(reason="broken until https://github.com/pola-rs/polars/issues/20162 is fixed")
def test_array_domain_query():
    pl = pytest.importorskip("polars")

    # this triggers construction of a lazyframe domain from the schema
    context = dp.Context.compositor(
        data=pl.LazyFrame(pl.Series("alpha", [["A", "B", "C"]] * 100, dtype=pl.Array(pl.String, 3))),
        privacy_unit=dp.unit_of(contributions=1),
        privacy_loss=dp.loss_of(epsilon=1.0, delta=1e-7),
        split_evenly_over=1,
    )
    
    # this is broken until https://github.com/pola-rs/polars/issues/20162 is fixed
    context.query().with_columns(pl.col.alpha.explode()).select(dp.len()).release().collect()

def test_arithmetic():
    pl = pytest.importorskip("polars")

    context = dp.Context.compositor(
        data=pl.LazyFrame({
            "data": [1, 2, 3] * 100,
            "weights": [0.2, 0.5, 0.7] * 100,
        }),
        privacy_unit=dp.unit_of(contributions=1),
        privacy_loss=dp.loss_of(epsilon=1.0),
        split_evenly_over=1,
        margins=[dp.polars.Margin(by=(), max_length=300)],
    )

    context.query().filter(pl.col.data.truediv(2) > 1.5).select(dp.len()).summarize()

    observed = (
        context.query()
        .select((pl.col.data * pl.col.weights).fill_null(0).fill_nan(0).dp.sum((0, 5)))
        .release()
        .collect()["data"][0]
    )

    # expectation is 330.0
    assert 260 < observed < 400

@pytest.mark.parametrize("keep", ["first", "last", "sample"])
def test_truncate_per_group(keep):
    pl = pytest.importorskip("polars")

    context = dp.Context.compositor(
        data=pl.LazyFrame({"alpha": ["A", "B", "C"] * 100, "id": [1, 2, 3] * 100}),
        privacy_unit=dp.unit_of(contributions=1, identifier="id"),
        privacy_loss=dp.loss_of(epsilon=1.0, delta=1e-7),
        split_evenly_over=1,
    )

    query = context.query().truncate_per_group(2, keep=keep).select(dp.len())
    assert query.summarize()["scale"][0] == 2.0000000000000004  # type: ignore[index]

    context = dp.Context.compositor(
        data=pl.LazyFrame({"alpha": ["A", "B", "C", "D"] * 100, "id": list(range(100)) * 4}),
        privacy_unit=dp.unit_of(
            contributions=[
                Bound(
                    by=["alpha"],
                    num_groups=1,
                    per_group=1,
                )
            ],
            identifier="id",
        ),
        privacy_loss=dp.loss_of(rho=0.5, delta=1e-7),
        split_evenly_over=1,
    )

    query = context.query().truncate_per_group(2, by=["alpha"], keep=keep).group_by("alpha").agg(dp.len())
    assert query.summarize()["scale"][0] == 2.0000000000000004  # type: ignore[index]


def test_truncate_per_group_sort_by():
    pl = pytest.importorskip("polars")

    context = dp.Context.compositor(
        data=pl.LazyFrame({"alpha": ["A", "B", "C"] * 100, "id": [1, 2, 3] * 100, "sort": [3, 2, 1] * 100}),
        privacy_unit=dp.unit_of(contributions=1, identifier="id"),
        privacy_loss=dp.loss_of(epsilon=1.0, delta=1e-7),
        split_evenly_over=1,
    )

    query = context.query().truncate_per_group(2, keep=dp.polars.SortBy(pl.col("sort"))).select(dp.len())
    assert query.summarize()["scale"][0] == 2.0000000000000004  # type: ignore[index]


def test_truncate_error_messages():
    pl = pytest.importorskip("polars")

    context = dp.Context.compositor(
        data=pl.LazyFrame({"alpha": ["A", "B", "C"] * 100, "id": [1, 2, 3] * 100, "sort": [3, 2, 1] * 100}),
        privacy_unit=dp.unit_of(contributions=1, identifier="id"),
        privacy_loss=dp.loss_of(epsilon=1.0, delta=1e-7),
        split_evenly_over=1,
    )

    query = context.query().truncate_num_groups(1, by=["alpha"]).select(dp.len())
    with pytest.raises(dp.OpenDPException, match="`per_group` contributions is unknown. This is likely due to a missing truncation"):
        query.summarize()
    
    query = context.query().truncate_num_groups(1, by=["alpha"]).group_by("sort").agg(dp.len())
    with pytest.raises(dp.OpenDPException, match=re.escape('To bound `num_groups` in the Context API, try using `.truncate_num_groups(num_groups, by=[col("sort")])`. To bound `per_group` in the Context API, try using `.truncate_per_group(per_group, by=[col("sort")])`.')):
        query.summarize()


@pytest.mark.parametrize("keep", ["first", "last"])
def test_truncate_num_groups(keep):
    pl = pytest.importorskip("polars")

    context = dp.Context.compositor(
        data=pl.LazyFrame({"alpha": ["A", "B", "C"] * 100, "id": [1, 2, 3] * 100}),
        privacy_unit=dp.unit_of(contributions=1, identifier="id"),
        privacy_loss=dp.loss_of(epsilon=1.0, delta=1e-7),
        split_evenly_over=1,
    )

    query = (
        context.query()
        .truncate_per_group(2)
        .truncate_num_groups(1, keep=keep, by=["alpha"])
        .group_by("alpha")
        .agg(dp.len())
    )
    assert query.summarize()["scale"][0] == 2.0000000000000004  # type: ignore[index]


@pytest.mark.parametrize("privacy_unit", [dp.unit_of(changes=1), dp.unit_of(changes=1, ordered=True)])
def test_private_lazyframe_bounded_dp(privacy_unit):
    pl = pytest.importorskip("polars")

    context = dp.Context.compositor(
        data=pl.LazyFrame({"alpha": ["A", "B", "C"] * 100, "id": [1, 2, 3] * 100}),
        privacy_unit=privacy_unit,
        privacy_loss=dp.loss_of(epsilon=1.0, delta=1e-7),
        split_evenly_over=1,
        margins=[dp.polars.Margin(by=(), max_length=300)],
    )

    query = context.query().select(pl.col.id.fill_null(0).dp.sum((0, 3)))
    assert query.summarize()["scale"][0] == 3.000000000000001  # type: ignore[index]

def test_lazyframe_bounded_dp_truncation():
    pl = pytest.importorskip("polars")

    context = dp.Context.compositor(
        data=pl.LazyFrame({"alpha": ["A", "B", "C"] * 100, "id": [1, 2, 3] * 100}),
        privacy_unit=dp.unit_of(changes=1, identifier="id"),
        privacy_loss=dp.loss_of(epsilon=1.0, delta=1e-7),
        split_evenly_over=1,
    )

    # the unused with_columns is for test coverage of .truncate, 
    # which retrieves input domain from prior query
    query = context.query().with_columns(x=pl.lit(10)).truncate_per_group(3).select(dp.len())
    assert query.summarize()["scale"][0] == 6.000000000000001  # type: ignore[index]


def test_unnecessary_lazyframe_truncation():
    pl = pytest.importorskip("polars")

    context = dp.Context.compositor(
        data=pl.LazyFrame({"alpha": ["A", "B", "C"] * 100, "id": [1, 2, 3] * 100}),
        privacy_unit=dp.unit_of(contributions=1),
        privacy_loss=dp.loss_of(epsilon=1.0, delta=1e-7),
        split_evenly_over=1,
        margins=[dp.polars.Margin(by=(), max_length=300)],
    )

    with pytest.raises(ValueError, match="truncation is only valid when"):
        context.query().truncate_per_group(3)
    with pytest.raises(ValueError, match="truncation is only valid when"):
        context.query().truncate_num_groups(3, by=["alpha"])


def test_frame_distance():
    pl = pytest.importorskip("polars")

    context = dp.Context.compositor(
        data=pl.LazyFrame({"alpha": ["A", "B", "C"] * 100, "id": range(300)}),
        privacy_unit=dp.unit_of(contributions=[
            dp.polars.Bound(per_group=2)
        ], identifier="id"),
        privacy_loss=dp.loss_of(epsilon=1.0, delta=1e-8),
        split_evenly_over=1,
    )

    query = (
        context.query()
        # user can contribute one record per id per group (2 records per group)
        .truncate_per_group(1, by=["alpha"])
        # user can contribute one group per id (2 groups total)
        .truncate_num_groups(1, by=["alpha"])
        .group_by("alpha")
        .agg(dp.len())
    )
    # ...therefore sensitivity of count is 2 * 2
    assert query.summarize()["scale"][0] == 4.000000000000001  # type: ignore[index]


def test_zero_budget():
    pl = pytest.importorskip("polars")

    context = dp.Context.compositor(
        data=pl.LazyFrame({"alpha": ["A", "B", "C"] * 100, "id": [1, 2, 3] * 100}),
        privacy_unit=dp.unit_of(changes=1),
        privacy_loss=dp.loss_of(epsilon=1.0),
        split_evenly_over=1,
    )
    context.query().select(pl.len()).release().collect()



'''
This module requires extra installs: ``pip install 'opendp[mbi]'``

``mbi`` is short for "marginal-based inference", 
and is the name of the `Private-PGM <https://github.com/ryan112358/private-pgm>`_ package.
OpenDP uses Private-PGM to postprocess DP releases made with the OpenDP Library.

For convenience, all the members of this module are also available from :py:mod:`opendp.prelude`.
We suggest importing under the conventional name ``dp``:

.. code:: pycon

    >>> import opendp.prelude as dp

The members of this module will then be accessible at ``dp.mbi``.
'''

from ._aim import AIM
from ._fixed import Fixed
from ._mst import MST
from ._sequential import Sequential
from ._table import make_contingency_table, ContingencyTable
from ._utilities import Algorithm, Count, mirror_descent

__all__ = [
    "AIM",
    "Fixed",
    "MST",
    "Sequential",
    "make_contingency_table",
    "ContingencyTable",
    "Algorithm",
    "Count",
    "mirror_descent",
]



# type: ignore
def make_randomized_response_bool(prob: f64, constant_time: bool):
    input_domain = AtomDomain(bool)
    input_metric = DiscreteMetric()
    output_measure = MaxDivergence()
    
    if (prob < 0.5 or prob > 1):  # |\label{line:range}|
        raise Exception("probability must be in [0.5, 1]")

    if prob == 1.0:
        c = float("inf")
    else:
        c = prob.inf_div((1).neg_inf_sub(prob)).inf_ln()
    
    def privacy_map(d_in: u32) -> f64:  # |\label{line:map}|
        if d_in == 0:
            return 0
        else:
            return c

    def function(arg: bool) -> bool:  # |\label{line:fn}|
        sample = not sample_bernoulli_float(prob, constant_time)

        return arg ^ sample
    
    return Measurement(input_domain, function, input_metric, output_measure, privacy_map)


# type: ignore
def truncate_id_bound(
    id_bound: Bound,
    truncation: Bound,
    total_ids: Optional[int],
) -> Bound:
    # Once truncated, max contributions when grouped by "over" are bounded
    row_bound = Bound.by(truncation.by)

    # In each group, the worst-case row contributions is the
    # the number of ids contributed (known from id_bound)
    # times the number of rows contributed under each id (known from truncation),
    num_ids, num_rows = id_bound.per_group, truncation.per_group
    if num_ids is not None and num_rows is not None:
        row_bound = row_bound.with_per_group(num_ids.inf_mul(num_rows)) # |\label{line:per-group}|

    # Worst case number of groups contributed is the
    # total number of ids contributed (total_ids)
    # times the number of groups contributed under each id (known from truncation).
    num_groups_via_truncation = None  # |\label{line:num-groups-via-truncation}|
    if total_ids is not None and truncation.num_groups is not None:
        num_groups_via_truncation = total_ids.inf_mul(truncation.num_groups)

    # Alternatively, the number of groups contributed may be known outright from id_bound.
    # Use the smaller of the two if both are known.
    num_groups = option_min(num_groups_via_truncation, id_bound.num_groups)
    if num_groups is not None:
        row_bound = row_bound.with_num_groups(num_groups) # |\label{line:num-groups}|

    return row_bound



# Auto-generated. Do not edit!
'''
The ``measures`` module provides functions that measure the distance between probability distributions.
For more context, see :ref:`measures in the User Guide <measures-user-guide>`.

For convenience, all the functions of this module are also available from :py:mod:`opendp.prelude`.
We suggest importing under the conventional name ``dp``:

.. code:: pycon

    >>> import opendp.prelude as dp
'''
from deprecated.sphinx import deprecated # noqa: F401 (Not every file actually has deprecated functions.)

from opendp._convert import *
from opendp._lib import *
from opendp.mod import *
from opendp.typing import *
from opendp.typing import _substitute # noqa: F401

__all__ = [
    "_approximate_divergence_get_inner_measure",
    "_measure_equal",
    "_measure_free",
    "approximate",
    "fixed_smoothed_max_divergence",
    "max_divergence",
    "measure_debug",
    "measure_distance_type",
    "measure_type",
    "new_privacy_profile",
    "renyi_divergence",
    "smoothed_max_divergence",
    "user_divergence",
    "zero_concentrated_divergence"
]


def _approximate_divergence_get_inner_measure(
    privacy_measure: Measure
) -> Measure:
    r"""Retrieve the inner privacy measure of an approximate privacy measure.

    .. end-markdown

    :param privacy_measure: The privacy measure to inspect
    :type privacy_measure: Measure
    :raises TypeError: if an argument's type differs from the expected type
    :raises UnknownTypeException: if a type argument fails to parse
    :raises OpenDPException: packaged error from the core OpenDP library
    """
    # No type arguments to standardize.
    # Convert arguments to c types.
    c_privacy_measure = py_to_c(privacy_measure, c_type=Measure, type_name="AnyMeasure")

    # Call library function.
    lib_function = lib.opendp_measures___approximate_divergence_get_inner_measure
    lib_function.argtypes = [Measure]
    lib_function.restype = FfiResult

    output = c_to_py(unwrap(lib_function(c_privacy_measure), Measure))
    try:
        output.__opendp_dict__ = {
            '__function__': '_approximate_divergence_get_inner_measure',
            '__module__': 'measures',
            '__kwargs__': {
                'privacy_measure': privacy_measure
            },
        }
    except AttributeError:  # pragma: no cover
        pass
    return output


def _measure_equal(
    left: Measure,
    right: Measure
) -> bool:
    r"""Check whether two measures are equal.

    .. end-markdown

    :param left: Measure to compare.
    :type left: Measure
    :param right: Measure to compare.
    :type right: Measure
    :raises TypeError: if an argument's type differs from the expected type
    :raises UnknownTypeException: if a type argument fails to parse
    :raises OpenDPException: packaged error from the core OpenDP library
    """
    # No type arguments to standardize.
    # Convert arguments to c types.
    c_left = py_to_c(left, c_type=Measure, type_name="AnyMeasure")
    c_right = py_to_c(right, c_type=Measure, type_name="AnyMeasure")

    # Call library function.
    lib_function = lib.opendp_measures___measure_equal
    lib_function.argtypes = [Measure, Measure]
    lib_function.restype = FfiResult

    output = c_to_py(unwrap(lib_function(c_left, c_right), BoolPtr))
    try:
        output.__opendp_dict__ = {
            '__function__': '_measure_equal',
            '__module__': 'measures',
            '__kwargs__': {
                'left': left, 'right': right
            },
        }
    except AttributeError:  # pragma: no cover
        pass
    return output


def _measure_free(
    this
):
    r"""Internal function. Free the memory associated with `this`.

    .. end-markdown

    :param this: 
    :type this: Measure
    :raises TypeError: if an argument's type differs from the expected type
    :raises UnknownTypeException: if a type argument fails to parse
    :raises OpenDPException: packaged error from the core OpenDP library
    """
    # No type arguments to standardize.
    # Convert arguments to c types.
    c_this = this

    # Call library function.
    lib_function = lib.opendp_measures___measure_free
    lib_function.argtypes = [Measure]
    lib_function.restype = FfiResult

    output = c_to_py(unwrap(lib_function(c_this), ctypes.c_void_p))
    try:
        output.__opendp_dict__ = {
            '__function__': '_measure_free',
            '__module__': 'measures',
            '__kwargs__': {
                'this': this
            },
        }
    except AttributeError:  # pragma: no cover
        pass
    return output


def approximate(
    measure: Measure
) -> ApproximateDivergence:
    r"""Privacy measure used to define $\delta$-approximate PM-differential privacy.

    In the following definition, $d$ corresponds to privacy parameters $(d', \delta)$
    when also quantified over all adjacent datasets
    ($d'$ is the privacy parameter corresponding to privacy measure PM).
    That is, $(d', \delta)$ is no smaller than $d$ (by product ordering),
    over all pairs of adjacent datasets $x, x'$ where $Y \sim M(x)$, $Y' \sim M(x')$.
    $M(\cdot)$ is a measurement (commonly known as a mechanism).
    The measurement's input metric defines the notion of adjacency,
    and the measurement's input domain defines the set of possible datasets.

    [approximate in Rust documentation.](https://docs.rs/opendp/0.14.0/opendp/measures/fn.approximate.html)

    **Proof Definition:**

    For any two distributions $Y, Y'$ and 2-tuple $d = (d', \delta)$,
    where $d'$ is the distance with respect to privacy measure PM,
    $Y, Y'$ are $d$-close under the approximate PM measure whenever,
    for any choice of $\delta \in [0, 1]$,
    there exist events $E$ (depending on $Y$) and $E'$ (depending on $Y'$)
    such that $\Pr[E] \ge 1 - \delta$, $\Pr[E'] \ge 1 - \delta$, and

    $D_{\mathrm{PM}}^\delta(Y|_E, Y'|_{E'}) = D_{\mathrm{PM}}(Y|_E, Y'|_{E'})$

    where $Y|_E$ denotes the distribution of $Y$ conditioned on the event $E$.

    Note that this $\delta$ is not privacy parameter $\delta$ until quantified over all adjacent datasets,
    as is done in the definition of a measurement.

    .. end-markdown

    :param measure: inner privacy measure
    :type measure: Measure
    :raises TypeError: if an argument's type differs from the expected type
    :raises UnknownTypeException: if a type argument fails to parse
    :raises OpenDPException: packaged error from the core OpenDP library
    """
    # No type arguments to standardize.
    # Convert arguments to c types.
    c_measure = py_to_c(measure, c_type=Measure, type_name=None)

    # Call library function.
    lib_function = lib.opendp_measures__approximate
    lib_function.argtypes = [Measure]
    lib_function.restype = FfiResult

    output = c_to_py(unwrap(lib_function(c_measure), Measure))
    try:
        output.__opendp_dict__ = {
            '__function__': 'approximate',
            '__module__': 'measures',
            '__kwargs__': {
                'measure': measure
            },
        }
    except AttributeError:  # pragma: no cover
        pass
    return output


def fixed_smoothed_max_divergence(

) -> Measure:
    r"""Privacy measure used to define $(\epsilon, \delta)$-approximate differential privacy.

    In the following definition, $d$ corresponds to $(\epsilon, \delta)$ when also quantified over all adjacent datasets.
    That is, $(\epsilon, \delta)$ is no smaller than $d$ (by product ordering),
    over all pairs of adjacent datasets $x, x'$ where $Y \sim M(x)$, $Y' \sim M(x')$.
    $M(\cdot)$ is a measurement (commonly known as a mechanism).
    The measurement's input metric defines the notion of adjacency,
    and the measurement's input domain defines the set of possible datasets.

    **Proof Definition:**

    For any two distributions $Y, Y'$ and any 2-tuple $d$ of non-negative numbers $\epsilon$ and $\delta$,
    $Y, Y'$ are $d$-close under the fixed smoothed max divergence measure whenever

    $D_\infty^\delta(Y, Y') = \max_{S \subseteq \textrm{Supp}(Y)} \Big[\ln \dfrac{\Pr[Y \in S] + \delta}{\Pr[Y' \in S]} \Big] \leq \epsilon$.

    Note that this $\epsilon$ and $\delta$ are not privacy parameters $\epsilon$ and $\delta$ until quantified over all adjacent datasets,
    as is done in the definition of a measurement.

    .. end-markdown


    :raises TypeError: if an argument's type differs from the expected type
    :raises UnknownTypeException: if a type argument fails to parse
    :raises OpenDPException: packaged error from the core OpenDP library
    """
    # No type arguments to standardize.
    # No arguments to convert to c types.
    # Call library function.
    lib_function = lib.opendp_measures__fixed_smoothed_max_divergence
    lib_function.argtypes = []
    lib_function.restype = FfiResult

    output = c_to_py(unwrap(lib_function(), Measure))
    try:
        output.__opendp_dict__ = {
            '__function__': 'fixed_smoothed_max_divergence',
            '__module__': 'measures',
            '__kwargs__': {
                
            },
        }
    except AttributeError:  # pragma: no cover
        pass
    return output


def max_divergence(

) -> Measure:
    r"""Privacy measure used to define $\epsilon$-pure differential privacy.

    In the following proof definition, $d$ corresponds to $\epsilon$ when also quantified over all adjacent datasets.
    That is, $\epsilon$ is the greatest possible $d$
    over all pairs of adjacent datasets $x, x'$ where $Y \sim M(x)$, $Y' \sim M(x')$.
    $M(\cdot)$ is a measurement (commonly known as a mechanism).
    The measurement's input metric defines the notion of adjacency,
    and the measurement's input domain defines the set of possible datasets.

    **Proof Definition:**

    For any two distributions $Y, Y'$ and any non-negative $d$,
    $Y, Y'$ are $d$-close under the max divergence measure whenever

    $D_\infty(Y, Y') = \max_{S \subseteq \textrm{Supp}(Y)} \Big[\ln \dfrac{\Pr[Y \in S]}{\Pr[Y' \in S]} \Big] \leq d$.

    .. end-markdown


    :raises TypeError: if an argument's type differs from the expected type
    :raises UnknownTypeException: if a type argument fails to parse
    :raises OpenDPException: packaged error from the core OpenDP library
    """
    # No type arguments to standardize.
    # No arguments to convert to c types.
    # Call library function.
    lib_function = lib.opendp_measures__max_divergence
    lib_function.argtypes = []
    lib_function.restype = FfiResult

    output = c_to_py(unwrap(lib_function(), Measure))
    try:
        output.__opendp_dict__ = {
            '__function__': 'max_divergence',
            '__module__': 'measures',
            '__kwargs__': {
                
            },
        }
    except AttributeError:  # pragma: no cover
        pass
    return output


def measure_debug(
    this: Measure
) -> str:
    r"""Debug a `measure`.

    .. end-markdown

    :param this: The measure to debug (stringify).
    :type this: Measure
    :raises TypeError: if an argument's type differs from the expected type
    :raises UnknownTypeException: if a type argument fails to parse
    :raises OpenDPException: packaged error from the core OpenDP library
    """
    # No type arguments to standardize.
    # Convert arguments to c types.
    c_this = py_to_c(this, c_type=Measure, type_name=None)

    # Call library function.
    lib_function = lib.opendp_measures__measure_debug
    lib_function.argtypes = [Measure]
    lib_function.restype = FfiResult

    output = c_to_py(unwrap(lib_function(c_this), ctypes.c_char_p))
    try:
        output.__opendp_dict__ = {
            '__function__': 'measure_debug',
            '__module__': 'measures',
            '__kwargs__': {
                'this': this
            },
        }
    except AttributeError:  # pragma: no cover
        pass
    return output


def measure_distance_type(
    this: Measure
) -> str:
    r"""Get the distance type of a `measure`.

    .. end-markdown

    :param this: The measure to retrieve the distance type from.
    :type this: Measure
    :raises TypeError: if an argument's type differs from the expected type
    :raises UnknownTypeException: if a type argument fails to parse
    :raises OpenDPException: packaged error from the core OpenDP library
    """
    # No type arguments to standardize.
    # Convert arguments to c types.
    c_this = py_to_c(this, c_type=Measure, type_name=None)

    # Call library function.
    lib_function = lib.opendp_measures__measure_distance_type
    lib_function.argtypes = [Measure]
    lib_function.restype = FfiResult

    output = c_to_py(unwrap(lib_function(c_this), ctypes.c_char_p))
    try:
        output.__opendp_dict__ = {
            '__function__': 'measure_distance_type',
            '__module__': 'measures',
            '__kwargs__': {
                'this': this
            },
        }
    except AttributeError:  # pragma: no cover
        pass
    return output


def measure_type(
    this: Measure
) -> str:
    r"""Get the type of a `measure`.

    .. end-markdown

    :param this: The measure to retrieve the type from.
    :type this: Measure
    :raises TypeError: if an argument's type differs from the expected type
    :raises UnknownTypeException: if a type argument fails to parse
    :raises OpenDPException: packaged error from the core OpenDP library
    """
    # No type arguments to standardize.
    # Convert arguments to c types.
    c_this = py_to_c(this, c_type=Measure, type_name=None)

    # Call library function.
    lib_function = lib.opendp_measures__measure_type
    lib_function.argtypes = [Measure]
    lib_function.restype = FfiResult

    output = c_to_py(unwrap(lib_function(c_this), ctypes.c_char_p))
    try:
        output.__opendp_dict__ = {
            '__function__': 'measure_type',
            '__module__': 'measures',
            '__kwargs__': {
                'this': this
            },
        }
    except AttributeError:  # pragma: no cover
        pass
    return output


def new_privacy_profile(
    curve
):
    r"""Construct a PrivacyProfile from a user-defined callback.


    Required features: `contrib`, `honest-but-curious`

    **Why honest-but-curious?:**

    The privacy profile should implement a well-defined $\delta(\epsilon)$ curve:

    * monotonically decreasing
    * rejects epsilon values that are less than zero or nan
    * returns delta values only within [0, 1]

    .. end-markdown

    :param curve: A privacy curve mapping epsilon to delta
    :raises TypeError: if an argument's type differs from the expected type
    :raises UnknownTypeException: if a type argument fails to parse
    :raises OpenDPException: packaged error from the core OpenDP library

    :example:

    >>> dp.enable_features("contrib", "honest-but-curious")
    >>> profile = dp.new_privacy_profile(lambda eps: 1.0 if eps < 0.5 else 1e-8)
    ...
    >>> # epsilon is not enough, so delta saturates to one
    >>> profile.delta(epsilon=0.499)
    1.0
    >>> # invert it, find the suitable epsilon at this delta
    >>> profile.epsilon(delta=1e-8)
    0.5
    >>> # insufficient delta results in infinite epsilon
    >>> profile.epsilon(delta=1e-9)
    inf


    """
    assert_features("contrib", "honest-but-curious")

    # No type arguments to standardize.
    # Convert arguments to c types.
    c_curve = py_to_c(curve, c_type=CallbackFnPtr, type_name="f64")

    # Call library function.
    lib_function = lib.opendp_measures__new_privacy_profile
    lib_function.argtypes = [CallbackFnPtr]
    lib_function.restype = FfiResult

    output = c_to_py(unwrap(lib_function(c_curve), AnyObjectPtr))
    try:
        output.__opendp_dict__ = {
            '__function__': 'new_privacy_profile',
            '__module__': 'measures',
            '__kwargs__': {
                'curve': curve
            },
        }
    except AttributeError:  # pragma: no cover
        pass
    return output


def renyi_divergence(

) -> Measure:
    r"""Privacy measure used to define $\epsilon(\alpha)$-Rényi differential privacy.

    In the following proof definition, $d$ corresponds to an RDP curve when also quantified over all adjacent datasets.
    That is, an RDP curve $\epsilon(\alpha)$ is no smaller than $d(\alpha)$ for any possible choices of $\alpha$,
    and over all pairs of adjacent datasets $x, x'$ where $Y \sim M(x)$, $Y' \sim M(x')$.
    $M(\cdot)$ is a measurement (commonly known as a mechanism).
    The measurement's input metric defines the notion of adjacency,
    and the measurement's input domain defines the set of possible datasets.

    **Proof Definition:**

    For any two distributions $Y, Y'$ and any curve $d$,
    $Y, Y'$ are $d$-close under the Rényi divergence measure if,
    for any given $\alpha \in (1, \infty)$,

    $D_\alpha(Y, Y') = \frac{1}{1 - \alpha} \mathbb{E}_{x \sim Y'} \Big[\ln \left( \dfrac{\Pr[Y = x]}{\Pr[Y' = x]} \right)^\alpha \Big] \leq d(\alpha)$

    Note that this $\epsilon$ and $\alpha$ are not privacy parameters $\epsilon$ and $\alpha$ until quantified over all adjacent datasets,
    as is done in the definition of a measurement.

    .. end-markdown


    :raises TypeError: if an argument's type differs from the expected type
    :raises UnknownTypeException: if a type argument fails to parse
    :raises OpenDPException: packaged error from the core OpenDP library
    """
    # No type arguments to standardize.
    # No arguments to convert to c types.
    # Call library function.
    lib_function = lib.opendp_measures__renyi_divergence
    lib_function.argtypes = []
    lib_function.restype = FfiResult

    output = c_to_py(unwrap(lib_function(), Measure))
    try:
        output.__opendp_dict__ = {
            '__function__': 'renyi_divergence',
            '__module__': 'measures',
            '__kwargs__': {
                
            },
        }
    except AttributeError:  # pragma: no cover
        pass
    return output


def smoothed_max_divergence(

) -> Measure:
    r"""Privacy measure used to define $\epsilon(\delta)$-approximate differential privacy.

    In the following proof definition, $d$ corresponds to a privacy profile when also quantified over all adjacent datasets.
    That is, a privacy profile $\epsilon(\delta)$ is no smaller than $d(\delta)$ for all possible choices of $\delta$,
    and over all pairs of adjacent datasets $x, x'$ where $Y \sim M(x)$, $Y' \sim M(x')$.
    $M(\cdot)$ is a measurement (commonly known as a mechanism).
    The measurement's input metric defines the notion of adjacency,
    and the measurement's input domain defines the set of possible datasets.

    The distance $d$ is of type PrivacyProfile, so it can be invoked with an $\epsilon$
    to retrieve the corresponding $\delta$.

    **Proof Definition:**

    For any two distributions $Y, Y'$ and any curve $d(\cdot)$,
    $Y, Y'$ are $d$-close under the smoothed max divergence measure whenever,
    for any choice of non-negative $\epsilon$, and $\delta = d(\epsilon)$,

    $D_\infty^\delta(Y, Y') = \max_{S \subseteq \textrm{Supp}(Y)} \Big[\ln \dfrac{\Pr[Y \in S] + \delta}{\Pr[Y' \in S]} \Big] \leq \epsilon$.

    Note that $\epsilon$ and $\delta$ are not privacy parameters $\epsilon$ and $\delta$ until quantified over all adjacent datasets,
    as is done in the definition of a measurement.

    .. end-markdown


    :raises TypeError: if an argument's type differs from the expected type
    :raises UnknownTypeException: if a type argument fails to parse
    :raises OpenDPException: packaged error from the core OpenDP library
    """
    # No type arguments to standardize.
    # No arguments to convert to c types.
    # Call library function.
    lib_function = lib.opendp_measures__smoothed_max_divergence
    lib_function.argtypes = []
    lib_function.restype = FfiResult

    output = c_to_py(unwrap(lib_function(), Measure))
    try:
        output.__opendp_dict__ = {
            '__function__': 'smoothed_max_divergence',
            '__module__': 'measures',
            '__kwargs__': {
                
            },
        }
    except AttributeError:  # pragma: no cover
        pass
    return output


def user_divergence(
    descriptor: str
) -> Measure:
    r"""Privacy measure with meaning defined by an OpenDP Library user (you).

    Any two instances of UserDivergence are equal if their string descriptors are equal.


    Required features: `honest-but-curious`

    **Why honest-but-curious?:**

    The essential requirement of a privacy measure is that it is closed under postprocessing.
    Your privacy measure `D` must satisfy that, for any pure function `f` and any two distributions `Y, Y'`, then $D(Y, Y') \ge D(f(Y), f(Y'))$.

    Beyond this, you should also consider whether your privacy measure can be used to provide meaningful privacy guarantees to your privacy units.

    .. end-markdown

    :param descriptor: A string description of the privacy measure.
    :type descriptor: str
    :raises TypeError: if an argument's type differs from the expected type
    :raises UnknownTypeException: if a type argument fails to parse
    :raises OpenDPException: packaged error from the core OpenDP library
    """
    assert_features("honest-but-curious")

    # No type arguments to standardize.
    # Convert arguments to c types.
    c_descriptor = py_to_c(descriptor, c_type=ctypes.c_char_p, type_name="String")

    # Call library function.
    lib_function = lib.opendp_measures__user_divergence
    lib_function.argtypes = [ctypes.c_char_p]
    lib_function.restype = FfiResult

    output = c_to_py(unwrap(lib_function(c_descriptor), Measure))
    try:
        output.__opendp_dict__ = {
            '__function__': 'user_divergence',
            '__module__': 'measures',
            '__kwargs__': {
                'descriptor': descriptor
            },
        }
    except AttributeError:  # pragma: no cover
        pass
    return output


def zero_concentrated_divergence(

) -> Measure:
    r"""Privacy measure used to define $\rho$-zero concentrated differential privacy.

    In the following proof definition, $d$ corresponds to $\rho$ when also quantified over all adjacent datasets.
    That is, $\rho$ is the greatest possible $d$
    over all pairs of adjacent datasets $x, x'$ where $Y \sim M(x)$, $Y' \sim M(x')$.
    $M(\cdot)$ is a measurement (commonly known as a mechanism).
    The measurement's input metric defines the notion of adjacency,
    and the measurement's input domain defines the set of possible datasets.

    **Proof Definition:**

    For any two distributions $Y, Y'$ and any non-negative $d$,
    $Y, Y'$ are $d$-close under the zero-concentrated divergence measure if,
    for every possible choice of $\alpha \in (1, \infty)$,

    $D_\alpha(Y, Y') = \frac{1}{1 - \alpha} \mathbb{E}_{x \sim Y'} \Big[\ln \left( \dfrac{\Pr[Y = x]}{\Pr[Y' = x]} \right)^\alpha \Big] \leq d \cdot \alpha$.

    .. end-markdown


    :raises TypeError: if an argument's type differs from the expected type
    :raises UnknownTypeException: if a type argument fails to parse
    :raises OpenDPException: packaged error from the core OpenDP library
    """
    # No type arguments to standardize.
    # No arguments to convert to c types.
    # Call library function.
    lib_function = lib.opendp_measures__zero_concentrated_divergence
    lib_function.argtypes = []
    lib_function.restype = FfiResult

    output = c_to_py(unwrap(lib_function(), Measure))
    try:
        output.__opendp_dict__ = {
            '__function__': 'zero_concentrated_divergence',
            '__module__': 'measures',
            '__kwargs__': {
                
            },
        }
    except AttributeError:  # pragma: no cover
        pass
    return output



# type: ignore
class FloatExpFamily:
    def make_noise(self, input_space) -> Measurement[AtomDomain[T], T, AbsoluteDistance[QI], MO]:
        t_vec = make_vec(input_space) # |\label{line:make-vec}|
        m_noise = self.make_noise(t_vec.output_space()) # |\label{line:make-noise}|

        return t_vec >> m_noise >> then_index_or_default(0) # |\label{line:post}|
    


import re
import pytest
import opendp.prelude as dp


@pytest.mark.xfail(raises=dp.UnknownTypeException)
def test_iterable_data():
    # Currently fails with:
    #   opendp.mod.UnknownTypeException: <class 'range'>
    # Possible resolution:
    #   The data kwarg accepts iterables.
    context = dp.Context.compositor(
        data=range(100),
        privacy_unit=dp.unit_of(contributions=1),
        privacy_loss=dp.loss_of(epsilon=1.0),
        split_evenly_over=1,
    )
    sum_query = context.query().clamp((1.0, 10.0)).sum()
    sum_query.laplace()


def test_int_data_laplace_param():
    context = dp.Context.compositor(
        data=[1, 2, 3, 4, 5],
        privacy_unit=dp.unit_of(contributions=1),
        privacy_loss=dp.loss_of(epsilon=1.0),
        split_evenly_over=1,
    )
    sum_query = context.query().clamp((1, 10)).sum()
    sum_query.laplace(100)


@pytest.mark.xfail(raises=dp.OpenDPException)
def test_mean_without_size():
    # Currently fails with:
    #   opendp.mod.OpenDPException:
    #     MakeTransformation("dataset size must be known. Either specify size in the input domain or use make_resize")
    # Possible resolution:
    #   Error message suggests fixes in terms on the new Context API.
    context = dp.Context.compositor(
        data=[1.0, 2.0, 3.0, 4.0, 5.0],
        privacy_unit=dp.unit_of(contributions=1),
        privacy_loss=dp.loss_of(epsilon=1.0),
        split_evenly_over=1,
    )
    mean_query = context.query().clamp((1.0, 10.0)).mean()
    mean_query.laplace()


@pytest.mark.xfail(raises=dp.OpenDPException)
def test_int_mean():
    # Currently fails with:
    #   opendp.mod.OpenDPException:
    #     FFI("No match for concrete type i32. You've got a debug binary! Debug binaries support fewer types. Consult https://docs.opendp.org/en/stable/contributing/development-environment.html#build-opendp")
    # Possible resolution:
    #   Should just be the same as any mean without a size.
    context = dp.Context.compositor(
        data=[1, 2, 3, 4, 5],
        privacy_unit=dp.unit_of(contributions=1),
        privacy_loss=dp.loss_of(epsilon=1.0),
        split_evenly_over=1,
    )
    mean_query = context.query().clamp((1, 10)).mean()
    mean_query.laplace()


def test_scalar_instead_of_vector():
    with pytest.raises(TypeError, match='To fix, wrap domain kwarg with dp.vector_domain()'):
        dp.Context.compositor(
            data=[1, 2, 3, 4, 5],
            privacy_unit=dp.unit_of(contributions=1),
            privacy_loss=dp.loss_of(epsilon=1.0),
            split_evenly_over=1,
            domain=dp.domain_of(int),
        )

def test_query_dir():
    context = dp.Context.compositor(
        data=[1, 2, 3, 4, 5],
        privacy_unit=dp.unit_of(contributions=1),
        privacy_loss=dp.loss_of(epsilon=1.0),
        split_evenly_over=1,
    )
    query_dir = dir(context.query())
    assert 'count' in query_dir
    assert 'laplace' in query_dir


def test_string_instead_of_tuple_for_margin_key():
    pl = pytest.importorskip("polars")

    lf = pl.LazyFrame(
        {"a_column": [1, 2, 3, 4]},
        schema={"a_column": pl.Int32},
    )

    with pytest.raises(ValueError, match=re.escape('by (a_column) must be a sequence type; Did you mean ["a_column"]?')):
        dp.Context.compositor(
            data=lf,
            privacy_unit=dp.unit_of(contributions=1),
            privacy_loss=dp.loss_of(epsilon=1.0),
            split_evenly_over=1,
            margins=[
                # To reproduce failure, the column name must be multiple characters.
                # TODO: We want to fail earlier because the key is not a tuple.
                # (mypy does catch this, so we need "type: ignore", but we can't rely on users running mypy.)
                dp.polars.Margin(by=("a_column"), invariant="keys", max_length=5), # type: ignore
            ],
        )


def test_margins_dict_instead_of_list():
    pl = pytest.importorskip("polars")

    lf = pl.LazyFrame(
        {"col": [1, 2, 3, 4]},
        schema={"col": pl.Int32},
    )

    with pytest.warns(DeprecationWarning, match=re.escape('Margin dicts should be replaced with lists, with the key supplied as the "by" kwarg')):
        dp.Context.compositor(
            data=lf,
            privacy_unit=dp.unit_of(contributions=1),
            privacy_loss=dp.loss_of(epsilon=1.0),
            split_evenly_over=1,
            margins={ # type: ignore[arg-type]
                ('col',): dp.polars.Margin(invariant="keys", max_length=5),
            }
        )


@pytest.mark.parametrize(
    "domain", [dp.lazyframe_domain([]), dp.series_domain("A", dp.atom_domain(T=bool))])
def test_polars_data_loader_error_is_human_readable(domain):
    pytest.importorskip("polars")
    overall_pipeline = dp.c.make_adaptive_composition(
        domain, dp.symmetric_distance(), dp.max_divergence(), d_in=1,
        d_mids=[1.])
    with pytest.raises(ValueError, match="expected Polars *"):
        overall_pipeline("I'm not the right type!")


def test_polars_expr_loader_error_is_human_readable():
    pl = pytest.importorskip("polars")
    with pytest.raises(ValueError, match="expected Polars Expr"):
        dp.m.make_private_expr(
            dp.wild_expr_domain([]),
            dp.symmetric_distance(),
            dp.max_divergence(),
            pl.LazyFrame({}),
        )


def test_unrecognized_column():
    pl = pytest.importorskip("polars")

    context = dp.Context.compositor(
        data=pl.LazyFrame({"A": [1, 2, 3, 4], "B": [5, 6, 7, 8]}),
        privacy_unit=dp.unit_of(contributions=1),
        privacy_loss=dp.loss_of(epsilon=1),
        split_evenly_over=1,
        margins=[]
    )
    config = pl.col("X").fill_null(0).dp.mean((0, 10))

    plain_query = context.query().select(config)
    with pytest.raises(dp.OpenDPException, match=r"unrecognized column 'X' in output domain; expected one of: A, B"):
        plain_query.release()

def test_without_max_partition_length():
    pl = pytest.importorskip("polars")

    context_wo_margin = dp.Context.compositor(
        data=pl.LazyFrame({"A": [1, 2, 3, 4], "B": [5, 6, 7, 8]}),
        privacy_unit=dp.unit_of(contributions=1),
        privacy_loss=dp.loss_of(epsilon=1),
        split_evenly_over=2,
        margins=[],
    )

    config = pl.col("A").fill_null(0).dp.mean((0, 10))

    plain_query = context_wo_margin.query().select(config)
    with pytest.raises(dp.OpenDPException, match=r"must specify 'max_length' in a margin with by=\[\]"):
        plain_query.release()

    agg_query = context_wo_margin.query().group_by(["B"]).agg(config)
    with pytest.raises(dp.OpenDPException, match="must specify 'max_length' in a margin with by=\\[col\\(\"B\"\\)\\]"):
        agg_query.release()
    
# Add a margin and try the same queries again:
def test_with_max_length():
    pl = pytest.importorskip("polars")

    context_w_margin = dp.Context.compositor(
        data=pl.LazyFrame({"A": [1, 2, 3, 4], "B": [5, 6, 7, 8]}),
        privacy_unit=dp.unit_of(contributions=1),
        privacy_loss=dp.loss_of(epsilon=1),
        split_evenly_over=2,
        margins=[dp.polars.Margin(
            by=[],
            max_length=1
        )],
    )

    config = pl.col("A").fill_null(0).dp.mean((0, 10))

    plain_query = context_w_margin.query().select(config)
    plain_query.release()

    agg_query = context_w_margin.query().group_by(["B"]).agg(config)
    with pytest.raises(dp.OpenDPException, match=re.escape('The key-set of {col("B")} is private')):
        agg_query.release()


# Make key set invariant:
def test_with_max_length_and_invariant_keys():
    pl = pytest.importorskip("polars")
    
    context_w_invariant_keys = dp.Context.compositor(
        data=pl.LazyFrame({"A": [1, 2, 3, 4], "B": [5, 6, 7, 8]}),
        privacy_unit=dp.unit_of(contributions=1),
        privacy_loss=dp.loss_of(epsilon=1),
        split_evenly_over=2,
        margins=[
            dp.polars.Margin(
                by=[],
                max_length=1
            ),
            dp.polars.Margin(
                by=['B'],
                max_length=1,
                invariant='keys',
            ),
        ],
    )

    config = pl.col("A").fill_null(0).dp.mean((0, 10))

    plain_query = context_w_invariant_keys.query().select(config)
    plain_query.release()

    agg_query = context_w_invariant_keys.query().group_by(["B"]).agg(config)
    agg_query.release()


# type: ignore
def exponential_top_k(x: list[TIA], scale: RBig, k: usize, negate: bool):
    sign = Sign.from_(negate)
    scale = scale.into_rational()

    y = [x_i.into_rational() * sign for x_i in x]  # `\label{negate}`
    return peel_permute_and_flip(y, scale, k)



# type: ignore
class CompositionMeasure(ZeroConcentratedDivergence):
    def composability(  # |\label{line:composability}|
        self, adaptivity: Adaptivity
    ) -> Composability:
        match adaptivity:
            case Adaptivity.FullyAdaptive:
                return Composability.Sequential
            case _:
                return Composability.Concurrent

    def compose(self, d_mids: Vec[Self_Distance]) -> Self_Distance:
        d_out = 0.0
        for d_mid in d_mids:
            d_out = d_out.inf_add(d_mid)
        return d_out



'''
The ``extras`` module contains pure-python DP tools.
These are not yet supported by the Rust or R bindings.
Some tools require extra, optional Python dependencies, and are grouped under corresponding namespaces.
'''
